# coding: utf-8

"""
    imandra-http-server

    Query Imandra via HTTP. See also https://github.com/aestheticIntegration/bs-imandra-client for a sample client implementation and OCaml API types.  # noqa: E501

    The version of the OpenAPI document: 0.0.1
    Generated by OpenAPI Generator (https://openapi-generator.tech)

    Do not edit the class manually.
"""


from __future__ import annotations
import pprint
import re  # noqa: F401
import json



from pydantic import BaseModel, Field, StrictStr

class WithUnknownReasonBody(BaseModel):
    """
    WithUnknownReasonBody
    """
    unknown_reason: StrictStr = Field(...)
    __properties = ["unknown_reason"]

    class Config:
        """Pydantic configuration"""
        allow_population_by_field_name = True
        validate_assignment = True

    def to_str(self) -> str:
        """Returns the string representation of the model using alias"""
        return pprint.pformat(self.dict(by_alias=True))

    def to_json(self) -> str:
        """Returns the JSON representation of the model using alias"""
        return json.dumps(self.to_dict())

    @classmethod
    def from_json(cls, json_str: str) -> WithUnknownReasonBody:
        """Create an instance of WithUnknownReasonBody from a JSON string"""
        return cls.from_dict(json.loads(json_str))

    def to_dict(self):
        """Returns the dictionary representation of the model using alias"""
        _dict = self.dict(by_alias=True,
                          exclude={
                          },
                          exclude_none=True)
        return _dict

    @classmethod
    def from_dict(cls, obj: dict) -> WithUnknownReasonBody:
        """Create an instance of WithUnknownReasonBody from a dict"""
        if obj is None:
            return None

        if not isinstance(obj, dict):
            return WithUnknownReasonBody.parse_obj(obj)

        _obj = WithUnknownReasonBody.parse_obj({
            "unknown_reason": obj.get("unknown_reason")
        })
        return _obj




import urllib
import urllib.request
import urllib.error
import urllib.parse
import json
import gzip
import io
import tarfile
import base64


def analyze(auth, job_file, path):
    with open(job_file, "r") as f:
        job_file_contents = f.read()

    bytes_io = io.BytesIO()
    tf = tarfile.open(mode="w", fileobj=bytes_io)
    tf.add(path, arcname="")

    b64_bytes = base64.b64encode(bytes_io.getvalue())
    b64_string = b64_bytes.decode("ascii")

    payload = {"job": job_file_contents, "pathTarArchiveBase64": b64_string}

    json_str = json.dumps(payload)
    json_bytes = json_str.encode("utf-8")

    content = gzip.compress(json_bytes)

    url = "{}/{}".format(auth.imandra_web_host, "api/imandra-cfb/analyze")

    clen = len(content)
    headers = {
        "X-Auth": auth.token,
        "Content-Length": clen,
        "Content-Encoding": "gzip",
        "Accept-Encoding": "gzip",
    }
    request = urllib.request.Request(url, content, headers=headers)

    try:
        with urllib.request.urlopen(request) as response:
            if response.headers["content-encoding"] == "gzip":
                return gzip.decompress(response.read()).decode("utf-8")
            else:
                return response.read().decode("utf-8")

    except urllib.error.HTTPError as e:
        raise ValueError(e.read().decode("utf-8"))



import subprocess
import atexit
import signal
import os
import os.path
import urllib
import urllib.parse
import urllib.request
import urllib.error
import tarfile
from pathlib import Path
import sys
import time
import socket
import ipaddress
import shutil
import filecmp

from .. import api


# https://stackoverflow.com/a/47919356/4956000
def is_loopback(host):
    for family in (socket.AF_INET, socket.AF_INET6):
        try:
            r = socket.getaddrinfo(host, None, family, socket.SOCK_STREAM)
        except socket.gaierror:
            return False
        for family, _, _, _, sockaddr in r:
            if not ipaddress.ip_address(sockaddr[0]).is_loopback:
                return False
    return True


def check_imandra_ready(port, timeout):
    request = urllib.request.Request(f"http://localhost:{port}/status")
    for x in range(timeout):
        time.sleep(1)
        try:
            response = urllib.request.urlopen(request)
            if response.read().decode("UTF-8") == "OK":
                return True
        except:
            pass
    return False


class IplLogAnalysis:
    def __init__(
        self,
        auth,
        ipl_file,
        traces,
        organization,
        callback,
        interactive,
        imandra_host,
        imandra_port,
        json_out,
        decomp_job_id,
        runner_image,
        sender_comp_id,
    ):
        self._auth = auth
        self._ipl_file = ipl_file
        self._traces_dir = Path(traces).absolute()
        self._organization = organization
        self._callback = callback
        self._interactive = interactive == "true"
        self._imandra_host = imandra_host
        self._imandra_port = imandra_port
        self._json_out = Path(json_out).absolute() if json_out is not None else None
        self._decomp_job_id = decomp_job_id
        self._runner_image = runner_image
        self._sender_comp_id = sender_comp_id

    def login(self):
        self._auth.login()

    def _start_imandra(self, timeout):
        self._imandra_host = "localhost"
        self._imandra_port = 3000
        args = [
            "imandra-http-api",
            "--skip-update",
            "--dir",
            self.job_model_path(),
        ]
        imandra_proc = subprocess.Popen(
            args,
            stdout=subprocess.PIPE,
            stderr=subprocess.STDOUT,
        )
        imandra_pid = imandra_proc.pid

        def kill_imandra():
            os.kill(imandra_pid, signal.SIGTERM)

        atexit.register(kill_imandra)
        self._imandra_proc = imandra_proc
        self._imandra_pid = imandra_pid

        return check_imandra_ready(self._imandra_port, timeout)

    def start_imandra(self, timeout):
        if self._imandra_host is not None or self._imandra_port is not None:
            self._imandra_host = (
                self._imandra_host if self._imandra_host is not None else "localhost"
            )
            self._imandra_port = (
                self._imandra_port if self._imandra_port is not None else 3000
            )
            return True
        else:
            return self._start_imandra(timeout)

    def submit_job(self):
        job_id = api.ipl.log_analysis_builder(
            self._auth,
            self._ipl_file,
            self._organization,
            self._callback,
            self._decomp_job_id,
        )
        self._job_id = job_id
        return job_id

    def use_job(self, uuid):
        self._job_id = uuid

    def data_dir(self):
        return os.path.join(self._auth.folder_path, "data")

    def job_data_dir(self):
        return os.path.join(self.data_dir(), self._job_id)

    def find_local_job(self):
        if not self._ipl_file:
            return None

        data_dir = self.data_dir()

        def is_candidate(p):
            return os.path.isdir(
                os.path.join(data_dir, p, "lib", "ipl-log-analysis")
            ) and os.path.isfile(os.path.join(data_dir, p, "model.ipl"))

        job_dirs = list(
            filter(
                is_candidate,
                os.listdir(data_dir),
            )
        )

        for d in job_dirs:
            model_file = os.path.join(data_dir, d, "model.ipl")
            if filecmp.cmp(model_file, self._ipl_file):
                return d

        return None

    def job_archive_path(self):
        return os.path.join(self.data_dir(), "{}.{}".format(self._job_id, "tar.gz"))

    def job_model_path(self):
        return os.path.join(self.job_data_dir(), "ipl", "gen", "model")

    def prepare_job_data(self):
        data_dir = self.data_dir()

        if not os.path.exists(data_dir):
            print(f"Creating directory {data_dir}...")
            os.mkdir(data_dir)

        job_archive_path = self.job_archive_path()
        if not os.path.exists(job_archive_path):
            response = api.ipl.data(self._auth, self._job_id)
            with open(job_archive_path, "wb") as data_file:
                data_file.write(response["content"])

        job_data_dir = self.job_data_dir()
        if not os.path.exists(job_data_dir):
            print(f"Extracting job data in {job_data_dir}...")
            os.mkdir(job_data_dir)

            tar = tarfile.open(job_archive_path, "r:gz")
            tar.extractall(job_data_dir)
            tar.close()

            response = api.ipl.data(self._auth, self._job_id, file="version")
            with open(f"{job_data_dir}/version", "wb") as version_file:
                version_file.write(response["content"])

            ipl_file_copy = os.path.join(job_data_dir, "model.ipl")
            shutil.copyfile(src=self._ipl_file, dst=ipl_file_copy)

    def fetch_runner_binary(self, version):
        runner_dir = os.path.join(self.data_dir(), "log_analysis_runner")
        if not os.path.exists(runner_dir):
            os.mkdir(runner_dir)

        runner_filename = f"runner-static-bytecode-linux-{version}.tar.gz"
        runner_path = os.path.join(runner_dir, runner_filename)
        url = f"https://storage.googleapis.com/ipl-log-analysis-releases/{runner_filename}"

        request = urllib.request.Request(url)
        response = urllib.request.urlopen(request)
        content = response.read()

        with open(runner_path, "wb") as data_file:
            data_file.write(content)

        return runner_path

    def run_binary(self):
        job_data_dir = self.job_data_dir()

        with open(os.path.join(job_data_dir, "version"), "r") as version_file:
            job_runner_version = version_file.read().strip()

        try:
            runner_path = self.fetch_runner_binary(job_runner_version)
        except urllib.error.HTTPError as e:
            if e.code == 404:
                print(
                    f"WARNING: runner binary version {job_runner_version} not found. Using latest version..."
                )
                runner_path = self.fetch_runner_binary("latest")
            else:
                raise ValueError(e.read().decode("utf-8"))

        print("Extracting runner binary...")
        runner_tar = tarfile.open(runner_path, "r:gz")
        runner_tar.extractall(job_data_dir)
        runner_tar.close()

        installed_runner_path = os.path.join(job_data_dir, "bin", "ipl_log_analysis")
        run_args = [
            f"--imandra-port={self._imandra_port}",
            "--fix-traces",
            self._traces_dir,
        ]

        if self._interactive:
            run_args.append("--interactive=true")

        if self._json_out:
            run_args.append(f"--json-out={self._json_out}")

        if self._sender_comp_id:
            run_args.append(f"--sender-comp-id={self._sender_comp_id}")

        run_args.append(f"--imandra-host={self._imandra_host}")

        run_cmd = [installed_runner_path] + run_args
        self.run_analysis_process(run_cmd)

    def runner_image(self):
        if self._runner_image:
            print(f"Using runner image {self._runner_image}...")
            return self._runner_image
        else:
            env = os.getenv("IMANDRA_ENV", "prod")
            if env == "dev":
                runner_image = "europe-west1-docker.pkg.dev/imandra-dev/imandra/ipl-log-analysis-runner"
            else:
                runner_image = "imandra/ipl-log-analysis-runner"

            with open(
                os.path.join(self.job_data_dir(), "version"), "r"
            ) as version_file:
                job_runner_version = version_file.read().strip()
            tagged_runner_image = f"{runner_image}:{job_runner_version}"
            pull_p = subprocess.run(
                ["docker", "pull", tagged_runner_image],
                stdout=sys.stdout,
                stderr=subprocess.STDOUT,
            )
            if pull_p.returncode != 0:
                print(
                    f"WARNING: runner image {runner_image} with tag {job_runner_version} not found. Using latest version..."
                )
                tagged_runner_image = f"{runner_image}:latest"
                pull_p = subprocess.run(
                    ["docker", "pull", tagged_runner_image],
                    stdout=sys.stdout,
                    stderr=subprocess.STDOUT,
                )
                if pull_p.returncode != 0:
                    print("Failed to pull runner image")
                    return None
            return tagged_runner_image

    def run_docker(self):
        try:
            subprocess.run(["docker", "--version"], stdout=subprocess.DEVNULL)
        except FileNotFoundError as e:
            print(e)
            return

        runner_image = self.runner_image()
        if runner_image is None:
            sys.exit(1)

        runner_args_list = [
            f"--imandra-port={self._imandra_port}",
            "--fix-traces ../traces",
        ]
        if self._interactive:
            runner_args_list.append("--interactive true")

        if self._json_out is not None:
            runner_args_list.append("--json-out=out.json")

        if self._sender_comp_id:
            runner_args_list.append(f"--sender-comp-id={self._sender_comp_id}")

        if not is_loopback(self._imandra_host):
            runner_args_list.append(f"--imandra-host={self._imandra_host}")
        else:
            runner_args_list.append("--imandra-host=host.docker.internal")

        runner_args = " ".join(runner_args_list)
        runner_script = f"cd ipl-log-analysis && sudo tar zxf plugin.tar.gz && ./bin/ipl_log_analysis {runner_args}"

        if self._json_out is not None:
            # --json-out needs output file to exist
            os.close(os.open(self._json_out, os.O_CREAT))
            runner_script = "touch ipl-log-analysis/out.json && " + runner_script

        mount_points = [
            "-v",
            f"{self.job_archive_path()}:/home/ocaml/ipl-log-analysis/plugin.tar.gz",
            "-v",
            f"{self._traces_dir}:/home/ocaml/traces",
        ]
        if self._json_out is not None:
            mount_points = mount_points + [
                "-v",
                f"{self._json_out}:/home/ocaml/ipl-log-analysis/out.json",
            ]

        docker_args = mount_points + [
            "--platform=linux/amd64",
            "--rm",
            runner_image,
            "sh",
            "-uexc",
            runner_script,
        ]

        if self._interactive:
            docker_cmd = ["docker", "run"] + ["-i"] + docker_args
        else:
            docker_cmd = ["docker", "run"] + docker_args

        self.run_analysis_process(docker_cmd)

    def run_analysis_process(self, cmd):
        input_buffer = sys.stdin
        output_buffer = sys.stdout
        proc = subprocess.Popen(
            cmd,
            encoding="utf-8",
            stdin=subprocess.PIPE,
            stdout=output_buffer,
            universal_newlines=True,
            stderr=subprocess.STDOUT,
        )

        if self._interactive:
            while True:
                try:
                    inpt = input_buffer.readline()
                    print(inpt, file=proc.stdin, flush=True)
                except BrokenPipeError:
                    return
        else:
            while proc.poll() is None:
                time.sleep(0.5)



# coding: utf-8

"""
    imandra-http-server

    Query Imandra via HTTP. See also https://github.com/aestheticIntegration/bs-imandra-client for a sample client implementation and OCaml API types.  # noqa: E501

    The version of the OpenAPI document: 0.0.1
    Generated by OpenAPI Generator (https://openapi-generator.tech)

    Do not edit the class manually.
"""


import unittest
import datetime

import imandra_http_api_client
from imandra_http_api_client.models.instance_result import InstanceResult  # noqa: E501
from imandra_http_api_client.rest import ApiException

class TestInstanceResult(unittest.TestCase):
    """InstanceResult unit test stubs"""

    def setUp(self):
        pass

    def tearDown(self):
        pass

    def testInstanceResult(self):
        """Test InstanceResult"""
        # inst = InstanceResult()

if __name__ == '__main__':
    unittest.main()



from .api import ipl_common
from . import auth


class Client:
    def __init__(self, api_key=None, scheme=None, host=None, api_version=None):
        self._config = auth.Config(
            api_key=api_key, scheme=scheme, host=host, api_version=api_version
        )
        self._ipl_config: ipl_common.Config = {
            "host": self._config.get_url(),  # This includes scheme and API version path prefix.
            "headers": self._config.get_headers(),
        }

    def status(self, job_id):
        path = "ipl/jobs/{}/status"
        return ipl_common.status(self._ipl_config, path, job_id)

    def wait(self, job_id, interval=10):
        path = "ipl/jobs/{}/status"
        return ipl_common.wait(self._ipl_config, path, job_id, interval)

    def decompose(
        self, file, testgen_lang, organization, callback, doc_gen, parent_job_id
    ):
        path = "ipl/decompose"
        return ipl_common.decompose(
            self._ipl_config,
            path,
            file,
            testgen_lang,
            organization,
            callback,
            doc_gen,
            parent_job_id,
        )

    def data(self, job_id, file=None):
        path = "ipl/jobs/{}/data"
        return ipl_common.data(self._ipl_config, path, job_id, file)

    def simulator(self, zone, file):
        path = "ipl/simulator/create"
        return ipl_common.simulator(self._ipl_config, zone, path, file)

    def list_jobs(self, limit=10, job_type=None):
        path = "ipl/jobs"
        return ipl_common.list_jobs(self._ipl_config, path, limit, job_type)

    def unsat_analysis(self, file, organization, callback):
        path = "ipl/unsat-analysis/jobs"
        return ipl_common.unsat_analysis(
            self._ipl_config, path, file, organization, callback
        )

    def log_analysis_builder(
        self, file, organization=None, callback=None, decomp_job_id=None
    ):
        path = "ipl/log-analysis-builder/jobs"
        return ipl_common.log_analysis_builder(
            self._ipl_config, path, file, organization, callback, decomp_job_id
        )

    def cancel(self, job_id):
        path = "ipl/jobs/{}/cancel"
        return ipl_common.cancel(self._ipl_config, path, job_id)

    def validate(self, file, organization=None):
        path = "ipl/validate"
        return ipl_common.validate(self._ipl_config, path, file, organization)



import random
import uuid
import json
import os
import os.path

import urllib
import urllib.error
import urllib.parse
import urllib.request

import hashlib
import base64

import shutil
import webbrowser


_get_secrets_chars = b"abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789"


def _get_secrets():
    rchar = lambda: random.SystemRandom().choice(_get_secrets_chars)
    verifier = bytes(rchar() for _ in range(32))
    challenge = hashlib.sha256(verifier).digest()
    challenge = base64.b64encode(challenge, altchars=b"-_")
    challenge = challenge.replace(b"=", b"")
    return verifier, challenge


class Auth:
    """
    Auth encapsulates authentication information for Imandra's Cloud APIs at
    https://www.imandra.ai/api/. An instance of this class should be passed to
    other ``imandra`` library methods to make calls to Imandra's cloud APIs.
    """

    def __init__(self):
        """
        Construct a new Auth object. This constructor configures the object
        with the locations of the Imandra authentication on your local machine
        (e.g. from ``~/.imandra`` or `%APPDATA%\\imandra`), and the Imandra cloud API
        details.
        """

        # NOTE: config env var handling mimics imandra_network_client

        envs = {
            "dev": {
                "auth0_base_host": "test-ai.eu.auth0.com",
                "auth0_client_id": "DQs4kqaeTPAENZ8dAj64qEm2SbJgncNK",
                "auth0_audience": "https://www.dev.imandracapital.com/api",
                "imandra_web_host": "https://www.dev.imandracapital.com",
                "config_dir_name": "imandra-dev",
                "imandra_repl": "imandra-repl-dev",
                "gcloud_project": "imandra-dev",
            },
            "prod": {
                "auth0_base_host": "auth.imandra.ai",
                "auth0_client_id": "q2yGHBTLmJSia35zCOdkUpEecj9mQl6o",
                "auth0_audience": "https://www.imandra.ai/api",
                "imandra_web_host": "https://www.imandra.ai",
                "config_dir_name": "imandra",
                "imandra_repl": "imandra-repl",
                "gcloud_project": "imandra-prod",
            },
        }

        env = os.getenv("IMANDRA_ENV", "prod")
        self.gcloud_project = envs[env]["gcloud_project"]

        self.auth0_base_host = os.getenv(
            "IMANDRA_AUTH0_BASE_HOST", envs[env]["auth0_base_host"]
        )
        self.auth0_client_id = os.getenv(
            "IMANDRA_AUTH0_CLIENT_ID", envs[env]["auth0_client_id"]
        )
        self.auth0_audience = os.getenv(
            "IMANDRA_AUTH0_AUDIENCE", envs[env]["auth0_audience"]
        )
        self.imandra_web_host = os.getenv(
            "IMANDRA_WEB_HOST", envs[env]["imandra_web_host"]
        )
        self.imandra_repl = envs[env]["imandra_repl"]
        self.http_port = None

        if os.name == "nt":
            user_config_root = os.environ["APPDATA"]
            self.folder_path = os.path.join(
                user_config_root, envs[env]["config_dir_name"]
            )
        else:
            user_config_root = os.environ["HOME"]
            self.folder_path = os.path.join(
                user_config_root, "." + envs[env]["config_dir_name"]
            )

        self.redirect_uri = self.imandra_web_host + "/pkce-callback"
        if self.http_port:
            self.redirect_uri += "?redirect_port=" + str(self.http_port)

    def login(self):
        """
        Ensures that all information is present in order to make calls to the
        API. This method is best run interactively from `imandra-cli`, via
        `imandra auth login`.
        """
        self.ensure_folder()
        self.ensure_token()
        self.ensure_confirm()
        self.ensure_zone()

    def logout(self):
        """
        Remove all Imandra user configuration information on the local machine.
        """
        self.ensure_folder()
        shutil.rmtree(self.folder_path)

    def ensure_folder(self):
        """
        Ensures the existence of the local imandra configuration directory (`~/.imandra` or `%APPDATA%\\imandra`).
        """
        if not os.path.exists(self.folder_path):
            os.mkdir(self.folder_path)

        device_id_path = os.path.join(self.folder_path, "device_id")
        if not os.path.exists(device_id_path):
            self.device_id = str(uuid.uuid4())
            with open(device_id_path, "w") as device_id_file:
                device_id_file.write(self.device_id)
        else:
            with open(device_id_path, "r") as device_id_file:
                self.device_id = device_id_file.read()

    def ensure_token(self):
        """
        Ensures the existence of the local imandra login token. Prompts the user to log in if there isn't one.
        """
        token_path = os.path.join(self.folder_path, "login_token")
        if os.path.exists(token_path):
            with open(token_path, "r") as token_file:
                self.token = token_file.read().strip()
            return

        self.verifier, self.challenge = _get_secrets()
        link = self._make_link()
        print("Trying to open a browser window to authenticate you...")
        show_open_msg = False
        try:
            if not webbrowser.open(link):
                show_open_msg = True
        except webbrowser.Error:
            show_open_msg = True

        if show_open_msg:
            print(
                "We couldn't open the browser automatically! Please visit: \n{}".format(
                    link
                )
            )

        code_str = str(
            input("Once you've authenticated, please enter the code provided: ")
        ).strip()
        self._exchange_tokens(code_str, token_path)

    def ensure_zone(self):
        """
        Ensures the existence of the default zone for creation of Imandra core
        instances. Pick the zone closest to you from those available to reduce
        latency when interacting with Imandra Core instances.
        """
        zone_path = os.path.join(self.folder_path, "zone")
        if os.path.exists(zone_path):
            with open(zone_path, "r") as zone_file:
                self.zone = zone_file.read().strip()
            return

        path = "pod/clusters"
        url = "{}/{}".format(self.imandra_web_host, path)
        headers = {"X-Auth": self.token}

        request = urllib.request.Request(url, headers=headers)

        try:
            response = urllib.request.urlopen(request)
        except urllib.error.HTTPError as e:
            raise ValueError(e.read())
        j = json.loads(response.read())

        options = []
        option_strs = []
        for i, cluster in enumerate(j["clusters"]):
            options.append((i, cluster))
            option_strs.append("[{}] {}".format(str(i + 1), cluster))

        def ask_for_selection():
            query = input("Select cloud zone {} : ".format(" ".join(option_strs)))
            try:
                return int(query) - 1
            except ValueError:
                print("Not a valid selection")
                return ask_for_selection()

        selected_opt_idx = ask_for_selection()
        _, selected_cluster = options[selected_opt_idx]

        with open(zone_path, "w") as zone_file:
            zone_file.write(selected_cluster)

        self.zone = selected_cluster

    def _make_link(self):
        params = urllib.parse.urlencode(
            {
                "audience": self.auth0_audience,
                "response_type": "code",
                "client_id": self.auth0_client_id,
                "code_challenge": self.challenge,
                "code_challenge_method": "S256",
                "redirect_uri": self.redirect_uri,
            }
        )
        return "https://{}/{}?{}".format(self.auth0_base_host, "authorize", params)

    def _prompt_verify_email(self):
        query = input("Enter the verification code we sent to your email: ")
        evrid, token, intention = (
            base64.b64decode(query.encode(), altchars=b"-_").decode("utf8").split(",")
        )
        params = urllib.parse.urlencode(
            {"evrid": evrid, "token": token, "intention": intention}
        )
        url = "{}/{}?{}".format(self.imandra_web_host, "verify-email", params)
        headers = {"X-Device-Id": self.device_id}
        request = urllib.request.Request(url, headers=headers)
        try:
            _ = urllib.request.urlopen(request)
        except urllib.error.HTTPError as e:
            raise ValueError(e.read())

    def _exchange_tokens(self, code_str, token_path):
        tkn = {
            "grant_type": "authorization_code",
            "client_id": self.auth0_client_id,
            "code_verifier": self.verifier.decode("utf-8"),
            "code": code_str,
            "redirect_uri": self.redirect_uri,
        }
        url = "https://{}/{}".format(self.auth0_base_host, "oauth/token")
        data = json.dumps(tkn)
        clen = len(data)
        data = data.encode("utf-8")
        request = urllib.request.Request(
            url, data, {"Content-Type": "application/json", "Content-Length": str(clen)}
        )
        try:
            response = urllib.request.urlopen(request)
        except urllib.error.HTTPError as e:
            raise ValueError(e.read())
        auth_token = json.loads(response.read())
        headers = {
            "Authorization": "Bearer {}".format(auth_token["access_token"]),
            "X-Device-Id": self.device_id,
        }
        url = "{}/{}".format(self.imandra_web_host, "api/login-token-for")
        request = urllib.request.Request(url, headers=headers)
        try:
            response = urllib.request.urlopen(request)
            self.token = response.read().decode("utf-8")

            with open(token_path, "w") as token_file:
                token_file.write(self.token)

        except urllib.error.HTTPError as e:
            content = e.read().decode("utf-8")
            if e.code == 400 and content == "email-not-verified":
                self._prompt_verify_email()
                self.ensure_token()
            else:
                raise ValueError(content)

    def ensure_confirm(self):
        path = "api/user-info"
        url = "{}/{}".format(self.imandra_web_host, path)
        headers = {"X-Auth": self.token}

        request = urllib.request.Request(url, headers=headers)

        try:
            response = urllib.request.urlopen(request)
        except urllib.error.HTTPError as e:
            raise ValueError(e.read())
        j = json.loads(response.read())
        if not j["trial_activated"]:
            query = input("Do you want to start your Imandra trial? [Y/n] ")
            if query == "" or query[0].lower() == "y":
                path = "welcome/activate-trial"
                url = "{}/{}".format(self.imandra_web_host, path)
                request = urllib.request.Request(url, headers=headers)
                try:
                    response = urllib.request.urlopen(request)
                except urllib.error.HTTPError as e:
                    raise ValueError(e.read())
            else:
                raise ValueError("Trial not activated")

    def export(self):
        """
        Export your login details for use via another Imandra Cloud API client.
        """
        token_path = os.path.join(self.folder_path, "login_token")
        zone_path = os.path.join(self.folder_path, "zone")
        on_behalf_of_path = os.path.join(self.folder_path, "on_behalf_of")

        if not os.path.exists(token_path) or not os.path.exists(zone_path):
            raise ValueError("Not logged in")

        details = {}
        with open(token_path, "r") as token_file:
            details["auth"] = token_file.read().strip()

        with open(zone_path, "r") as zone_file:
            details["zone"] = zone_file.read().strip()

        if os.path.exists(on_behalf_of_path):
            with open(on_behalf_of_path, "r") as on_behalf_of_file:
                details["on_behalf_of"] = on_behalf_of_file.read().strip()
        else:
            details["on_behalf_of"] = None

        return details

    def import_(self, details):
        """
        Import your login details from stdin.
        """
        token_path = os.path.join(self.folder_path, "login_token")
        zone_path = os.path.join(self.folder_path, "zone")
        on_behalf_of_path = os.path.join(self.folder_path, "on_behalf_of")

        if details.get("auth"):
            self.ensure_folder()
            with open(token_path, "w") as f:
                f.write(details["auth"])

        if details.get("zone"):
            self.ensure_folder()
            with open(zone_path, "w") as f:
                f.write(details["zone"])

        if details.get("on_behalf_of"):
            self.ensure_folder()
            with open(on_behalf_of_path, "w") as f:
                f.write(details["zone"])



import inspect


def session():
    try:
        from .session import HttpInstanceSession
    except ModuleNotFoundError as err:
        note = """
            Install imandra with the optional 'http_api_client' dependency to enable imandra.session():

                pip install 'imandra[http_api_client]'
        """
        err.msg += "\n\n" + inspect.cleandoc(note)
        raise
    return HttpInstanceSession()



# coding: utf-8

"""
    imandra-http-server

    Query Imandra via HTTP. See also https://github.com/aestheticIntegration/bs-imandra-client for a sample client implementation and OCaml API types.  # noqa: E501

    The version of the OpenAPI document: 0.0.1
    Generated by OpenAPI Generator (https://openapi-generator.tech)

    Do not edit the class manually.
"""


import json
import pprint
import re  # noqa: F401
from aenum import Enum, no_arg





class MethodType(str, Enum):
    """
    MethodType
    """

    """
    allowed enum values
    """
    UNROLL = 'unroll'
    EXT_SOLVER = 'ext_solver'
    AUTO = 'auto'
    INDUCT = 'induct'

    @classmethod
    def from_json(cls, json_str: str) -> MethodType:
        """Create an instance of MethodType from a JSON string"""
        return MethodType(json.loads(json_str))





# coding: utf-8

"""
    imandra-http-server

    Query Imandra via HTTP. See also https://github.com/aestheticIntegration/bs-imandra-client for a sample client implementation and OCaml API types.  # noqa: E501

    The version of the OpenAPI document: 0.0.1
    Generated by OpenAPI Generator (https://openapi-generator.tech)

    Do not edit the class manually.
"""


import re  # noqa: F401
import io
import warnings

from pydantic import validate_arguments, ValidationError
from typing_extensions import Annotated

from imandra_http_api_client.models.decompose_request_src import DecomposeRequestSrc
from imandra_http_api_client.models.decompose_result import DecomposeResult
from imandra_http_api_client.models.eval_request_src import EvalRequestSrc
from imandra_http_api_client.models.eval_response import EvalResponse
from imandra_http_api_client.models.instance_request_name import InstanceRequestName
from imandra_http_api_client.models.instance_request_src import InstanceRequestSrc
from imandra_http_api_client.models.instance_response import InstanceResponse
from imandra_http_api_client.models.verify_request_name import VerifyRequestName
from imandra_http_api_client.models.verify_request_src import VerifyRequestSrc
from imandra_http_api_client.models.verify_response import VerifyResponse

from imandra_http_api_client.api_client import ApiClient
from imandra_http_api_client.api_response import ApiResponse
from imandra_http_api_client.exceptions import (  # noqa: F401
    ApiTypeError,
    ApiValueError
)


class DefaultApi(object):
    """NOTE: This class is auto generated by OpenAPI Generator
    Ref: https://openapi-generator.tech

    Do not edit the class manually.
    """

    def __init__(self, api_client=None):
        if api_client is None:
            api_client = ApiClient.get_default()
        self.api_client = api_client

    @validate_arguments
    def decompose(self, decompose_request_src : DecomposeRequestSrc, **kwargs) -> DecomposeResult:  # noqa: E501
        """Region decomposition of a function  # noqa: E501

          # noqa: E501
        This method makes a synchronous HTTP request by default. To make an
        asynchronous HTTP request, please pass async_req=True

        >>> thread = api.decompose(decompose_request_src, async_req=True)
        >>> result = thread.get()

        :param decompose_request_src: (required)
        :type decompose_request_src: DecomposeRequestSrc
        :param async_req: Whether to execute the request asynchronously.
        :type async_req: bool, optional
        :param _request_timeout: timeout setting for this request. If one
                                 number provided, it will be total request
                                 timeout. It can also be a pair (tuple) of
                                 (connection, read) timeouts.
        :return: Returns the result object.
                 If the method is called asynchronously,
                 returns the request thread.
        :rtype: DecomposeResult
        """
        kwargs['_return_http_data_only'] = True
        if '_preload_content' in kwargs:
            raise ValueError("Error! Please call the decompose_with_http_info method with `_preload_content` instead and obtain raw data from ApiResponse.raw_data")
        return self.decompose_with_http_info(decompose_request_src, **kwargs)  # noqa: E501

    @validate_arguments
    def decompose_with_http_info(self, decompose_request_src : DecomposeRequestSrc, **kwargs) -> ApiResponse:  # noqa: E501
        """Region decomposition of a function  # noqa: E501

          # noqa: E501
        This method makes a synchronous HTTP request by default. To make an
        asynchronous HTTP request, please pass async_req=True

        >>> thread = api.decompose_with_http_info(decompose_request_src, async_req=True)
        >>> result = thread.get()

        :param decompose_request_src: (required)
        :type decompose_request_src: DecomposeRequestSrc
        :param async_req: Whether to execute the request asynchronously.
        :type async_req: bool, optional
        :param _preload_content: if False, the ApiResponse.data will
                                 be set to none and raw_data will store the 
                                 HTTP response body without reading/decoding.
                                 Default is True.
        :type _preload_content: bool, optional
        :param _return_http_data_only: response data instead of ApiResponse
                                       object with status code, headers, etc
        :type _return_http_data_only: bool, optional
        :param _request_timeout: timeout setting for this request. If one
                                 number provided, it will be total request
                                 timeout. It can also be a pair (tuple) of
                                 (connection, read) timeouts.
        :param _request_auth: set to override the auth_settings for an a single
                              request; this effectively ignores the authentication
                              in the spec for a single request.
        :type _request_auth: dict, optional
        :type _content_type: string, optional: force content-type for the request
        :return: Returns the result object.
                 If the method is called asynchronously,
                 returns the request thread.
        :rtype: tuple(DecomposeResult, status_code(int), headers(HTTPHeaderDict))
        """

        _params = locals()

        _all_params = [
            'decompose_request_src'
        ]
        _all_params.extend(
            [
                'async_req',
                '_return_http_data_only',
                '_preload_content',
                '_request_timeout',
                '_request_auth',
                '_content_type',
                '_headers'
            ]
        )

        # validate the arguments
        for _key, _val in _params['kwargs'].items():
            if _key not in _all_params:
                raise ApiTypeError(
                    "Got an unexpected keyword argument '%s'"
                    " to method decompose" % _key
                )
            _params[_key] = _val
        del _params['kwargs']

        _collection_formats = {}

        # process the path parameters
        _path_params = {}

        # process the query parameters
        _query_params = []
        # process the header parameters
        _header_params = dict(_params.get('_headers', {}))
        # process the form parameters
        _form_params = []
        _files = {}
        # process the body parameter
        _body_params = None
        if _params['decompose_request_src'] is not None:
            _body_params = _params['decompose_request_src']

        # set the HTTP header `Accept`
        _header_params['Accept'] = self.api_client.select_header_accept(
            ['application/json'])  # noqa: E501

        # set the HTTP header `Content-Type`
        _content_types_list = _params.get('_content_type',
            self.api_client.select_header_content_type(
                ['application/json']))
        if _content_types_list:
                _header_params['Content-Type'] = _content_types_list

        # authentication setting
        _auth_settings = ['bearerAuth']  # noqa: E501

        _response_types_map = {
            '200': "DecomposeResult",
        }

        return self.api_client.call_api(
            '/decompose', 'POST',
            _path_params,
            _query_params,
            _header_params,
            body=_body_params,
            post_params=_form_params,
            files=_files,
            response_types_map=_response_types_map,
            auth_settings=_auth_settings,
            async_req=_params.get('async_req'),
            _return_http_data_only=_params.get('_return_http_data_only'),  # noqa: E501
            _preload_content=_params.get('_preload_content', True),
            _request_timeout=_params.get('_request_timeout'),
            collection_formats=_collection_formats,
            _request_auth=_params.get('_request_auth'))

    @validate_arguments
    def eval(self, eval_request_src : EvalRequestSrc, **kwargs) -> EvalResponse:  # noqa: E501
        """Evaluate a string of source code  # noqa: E501

          # noqa: E501
        This method makes a synchronous HTTP request by default. To make an
        asynchronous HTTP request, please pass async_req=True

        >>> thread = api.eval(eval_request_src, async_req=True)
        >>> result = thread.get()

        :param eval_request_src: (required)
        :type eval_request_src: EvalRequestSrc
        :param async_req: Whether to execute the request asynchronously.
        :type async_req: bool, optional
        :param _request_timeout: timeout setting for this request. If one
                                 number provided, it will be total request
                                 timeout. It can also be a pair (tuple) of
                                 (connection, read) timeouts.
        :return: Returns the result object.
                 If the method is called asynchronously,
                 returns the request thread.
        :rtype: EvalResponse
        """
        kwargs['_return_http_data_only'] = True
        if '_preload_content' in kwargs:
            raise ValueError("Error! Please call the eval_with_http_info method with `_preload_content` instead and obtain raw data from ApiResponse.raw_data")
        return self.eval_with_http_info(eval_request_src, **kwargs)  # noqa: E501

    @validate_arguments
    def eval_with_http_info(self, eval_request_src : EvalRequestSrc, **kwargs) -> ApiResponse:  # noqa: E501
        """Evaluate a string of source code  # noqa: E501

          # noqa: E501
        This method makes a synchronous HTTP request by default. To make an
        asynchronous HTTP request, please pass async_req=True

        >>> thread = api.eval_with_http_info(eval_request_src, async_req=True)
        >>> result = thread.get()

        :param eval_request_src: (required)
        :type eval_request_src: EvalRequestSrc
        :param async_req: Whether to execute the request asynchronously.
        :type async_req: bool, optional
        :param _preload_content: if False, the ApiResponse.data will
                                 be set to none and raw_data will store the 
                                 HTTP response body without reading/decoding.
                                 Default is True.
        :type _preload_content: bool, optional
        :param _return_http_data_only: response data instead of ApiResponse
                                       object with status code, headers, etc
        :type _return_http_data_only: bool, optional
        :param _request_timeout: timeout setting for this request. If one
                                 number provided, it will be total request
                                 timeout. It can also be a pair (tuple) of
                                 (connection, read) timeouts.
        :param _request_auth: set to override the auth_settings for an a single
                              request; this effectively ignores the authentication
                              in the spec for a single request.
        :type _request_auth: dict, optional
        :type _content_type: string, optional: force content-type for the request
        :return: Returns the result object.
                 If the method is called asynchronously,
                 returns the request thread.
        :rtype: tuple(EvalResponse, status_code(int), headers(HTTPHeaderDict))
        """

        _params = locals()

        _all_params = [
            'eval_request_src'
        ]
        _all_params.extend(
            [
                'async_req',
                '_return_http_data_only',
                '_preload_content',
                '_request_timeout',
                '_request_auth',
                '_content_type',
                '_headers'
            ]
        )

        # validate the arguments
        for _key, _val in _params['kwargs'].items():
            if _key not in _all_params:
                raise ApiTypeError(
                    "Got an unexpected keyword argument '%s'"
                    " to method eval" % _key
                )
            _params[_key] = _val
        del _params['kwargs']

        _collection_formats = {}

        # process the path parameters
        _path_params = {}

        # process the query parameters
        _query_params = []
        # process the header parameters
        _header_params = dict(_params.get('_headers', {}))
        # process the form parameters
        _form_params = []
        _files = {}
        # process the body parameter
        _body_params = None
        if _params['eval_request_src'] is not None:
            _body_params = _params['eval_request_src']

        # set the HTTP header `Accept`
        _header_params['Accept'] = self.api_client.select_header_accept(
            ['application/json'])  # noqa: E501

        # set the HTTP header `Content-Type`
        _content_types_list = _params.get('_content_type',
            self.api_client.select_header_content_type(
                ['application/json']))
        if _content_types_list:
                _header_params['Content-Type'] = _content_types_list

        # authentication setting
        _auth_settings = ['bearerAuth']  # noqa: E501

        _response_types_map = {
            '200': "EvalResponse",
            '422': "ResponseError",
        }

        return self.api_client.call_api(
            '/eval/by-src', 'POST',
            _path_params,
            _query_params,
            _header_params,
            body=_body_params,
            post_params=_form_params,
            files=_files,
            response_types_map=_response_types_map,
            auth_settings=_auth_settings,
            async_req=_params.get('async_req'),
            _return_http_data_only=_params.get('_return_http_data_only'),  # noqa: E501
            _preload_content=_params.get('_preload_content', True),
            _request_timeout=_params.get('_request_timeout'),
            collection_formats=_collection_formats,
            _request_auth=_params.get('_request_auth'))

    @validate_arguments
    def get_history(self, **kwargs) -> str:  # noqa: E501
        """Get a text version of the Logic History  # noqa: E501

        Gives an idea of what functions and theorems have been defined.  # noqa: E501
        This method makes a synchronous HTTP request by default. To make an
        asynchronous HTTP request, please pass async_req=True

        >>> thread = api.get_history(async_req=True)
        >>> result = thread.get()

        :param async_req: Whether to execute the request asynchronously.
        :type async_req: bool, optional
        :param _request_timeout: timeout setting for this request. If one
                                 number provided, it will be total request
                                 timeout. It can also be a pair (tuple) of
                                 (connection, read) timeouts.
        :return: Returns the result object.
                 If the method is called asynchronously,
                 returns the request thread.
        :rtype: str
        """
        kwargs['_return_http_data_only'] = True
        if '_preload_content' in kwargs:
            raise ValueError("Error! Please call the get_history_with_http_info method with `_preload_content` instead and obtain raw data from ApiResponse.raw_data")
        return self.get_history_with_http_info(**kwargs)  # noqa: E501

    @validate_arguments
    def get_history_with_http_info(self, **kwargs) -> ApiResponse:  # noqa: E501
        """Get a text version of the Logic History  # noqa: E501

        Gives an idea of what functions and theorems have been defined.  # noqa: E501
        This method makes a synchronous HTTP request by default. To make an
        asynchronous HTTP request, please pass async_req=True

        >>> thread = api.get_history_with_http_info(async_req=True)
        >>> result = thread.get()

        :param async_req: Whether to execute the request asynchronously.
        :type async_req: bool, optional
        :param _preload_content: if False, the ApiResponse.data will
                                 be set to none and raw_data will store the 
                                 HTTP response body without reading/decoding.
                                 Default is True.
        :type _preload_content: bool, optional
        :param _return_http_data_only: response data instead of ApiResponse
                                       object with status code, headers, etc
        :type _return_http_data_only: bool, optional
        :param _request_timeout: timeout setting for this request. If one
                                 number provided, it will be total request
                                 timeout. It can also be a pair (tuple) of
                                 (connection, read) timeouts.
        :param _request_auth: set to override the auth_settings for an a single
                              request; this effectively ignores the authentication
                              in the spec for a single request.
        :type _request_auth: dict, optional
        :type _content_type: string, optional: force content-type for the request
        :return: Returns the result object.
                 If the method is called asynchronously,
                 returns the request thread.
        :rtype: tuple(str, status_code(int), headers(HTTPHeaderDict))
        """

        _params = locals()

        _all_params = [
        ]
        _all_params.extend(
            [
                'async_req',
                '_return_http_data_only',
                '_preload_content',
                '_request_timeout',
                '_request_auth',
                '_content_type',
                '_headers'
            ]
        )

        # validate the arguments
        for _key, _val in _params['kwargs'].items():
            if _key not in _all_params:
                raise ApiTypeError(
                    "Got an unexpected keyword argument '%s'"
                    " to method get_history" % _key
                )
            _params[_key] = _val
        del _params['kwargs']

        _collection_formats = {}

        # process the path parameters
        _path_params = {}

        # process the query parameters
        _query_params = []
        # process the header parameters
        _header_params = dict(_params.get('_headers', {}))
        # process the form parameters
        _form_params = []
        _files = {}
        # process the body parameter
        _body_params = None
        # set the HTTP header `Accept`
        _header_params['Accept'] = self.api_client.select_header_accept(
            ['text/plain'])  # noqa: E501

        # authentication setting
        _auth_settings = ['bearerAuth']  # noqa: E501

        _response_types_map = {
            '200': "str",
        }

        return self.api_client.call_api(
            '/history', 'GET',
            _path_params,
            _query_params,
            _header_params,
            body=_body_params,
            post_params=_form_params,
            files=_files,
            response_types_map=_response_types_map,
            auth_settings=_auth_settings,
            async_req=_params.get('async_req'),
            _return_http_data_only=_params.get('_return_http_data_only'),  # noqa: E501
            _preload_content=_params.get('_preload_content', True),
            _request_timeout=_params.get('_request_timeout'),
            collection_formats=_collection_formats,
            _request_auth=_params.get('_request_auth'))

    @validate_arguments
    def get_status(self, **kwargs) -> str:  # noqa: E501
        """Get server status  # noqa: E501

        Can be used to check that the server is up and running.  # noqa: E501
        This method makes a synchronous HTTP request by default. To make an
        asynchronous HTTP request, please pass async_req=True

        >>> thread = api.get_status(async_req=True)
        >>> result = thread.get()

        :param async_req: Whether to execute the request asynchronously.
        :type async_req: bool, optional
        :param _request_timeout: timeout setting for this request. If one
                                 number provided, it will be total request
                                 timeout. It can also be a pair (tuple) of
                                 (connection, read) timeouts.
        :return: Returns the result object.
                 If the method is called asynchronously,
                 returns the request thread.
        :rtype: str
        """
        kwargs['_return_http_data_only'] = True
        if '_preload_content' in kwargs:
            raise ValueError("Error! Please call the get_status_with_http_info method with `_preload_content` instead and obtain raw data from ApiResponse.raw_data")
        return self.get_status_with_http_info(**kwargs)  # noqa: E501

    @validate_arguments
    def get_status_with_http_info(self, **kwargs) -> ApiResponse:  # noqa: E501
        """Get server status  # noqa: E501

        Can be used to check that the server is up and running.  # noqa: E501
        This method makes a synchronous HTTP request by default. To make an
        asynchronous HTTP request, please pass async_req=True

        >>> thread = api.get_status_with_http_info(async_req=True)
        >>> result = thread.get()

        :param async_req: Whether to execute the request asynchronously.
        :type async_req: bool, optional
        :param _preload_content: if False, the ApiResponse.data will
                                 be set to none and raw_data will store the 
                                 HTTP response body without reading/decoding.
                                 Default is True.
        :type _preload_content: bool, optional
        :param _return_http_data_only: response data instead of ApiResponse
                                       object with status code, headers, etc
        :type _return_http_data_only: bool, optional
        :param _request_timeout: timeout setting for this request. If one
                                 number provided, it will be total request
                                 timeout. It can also be a pair (tuple) of
                                 (connection, read) timeouts.
        :param _request_auth: set to override the auth_settings for an a single
                              request; this effectively ignores the authentication
                              in the spec for a single request.
        :type _request_auth: dict, optional
        :type _content_type: string, optional: force content-type for the request
        :return: Returns the result object.
                 If the method is called asynchronously,
                 returns the request thread.
        :rtype: tuple(str, status_code(int), headers(HTTPHeaderDict))
        """

        _params = locals()

        _all_params = [
        ]
        _all_params.extend(
            [
                'async_req',
                '_return_http_data_only',
                '_preload_content',
                '_request_timeout',
                '_request_auth',
                '_content_type',
                '_headers'
            ]
        )

        # validate the arguments
        for _key, _val in _params['kwargs'].items():
            if _key not in _all_params:
                raise ApiTypeError(
                    "Got an unexpected keyword argument '%s'"
                    " to method get_status" % _key
                )
            _params[_key] = _val
        del _params['kwargs']

        _collection_formats = {}

        # process the path parameters
        _path_params = {}

        # process the query parameters
        _query_params = []
        # process the header parameters
        _header_params = dict(_params.get('_headers', {}))
        # process the form parameters
        _form_params = []
        _files = {}
        # process the body parameter
        _body_params = None
        # set the HTTP header `Accept`
        _header_params['Accept'] = self.api_client.select_header_accept(
            ['text/plain'])  # noqa: E501

        # authentication setting
        _auth_settings = ['bearerAuth']  # noqa: E501

        _response_types_map = {
            '200': "str",
        }

        return self.api_client.call_api(
            '/status', 'GET',
            _path_params,
            _query_params,
            _header_params,
            body=_body_params,
            post_params=_form_params,
            files=_files,
            response_types_map=_response_types_map,
            auth_settings=_auth_settings,
            async_req=_params.get('async_req'),
            _return_http_data_only=_params.get('_return_http_data_only'),  # noqa: E501
            _preload_content=_params.get('_preload_content', True),
            _request_timeout=_params.get('_request_timeout'),
            collection_formats=_collection_formats,
            _request_auth=_params.get('_request_auth'))

    @validate_arguments
    def instance_by_name(self, instance_request_name : InstanceRequestName, **kwargs) -> InstanceResponse:  # noqa: E501
        """Search for an instance from a property (boolean function) by name. (The name should exist from a previous call to `/eval`).  This is the more 'primitive' call - calls to /*/by-src are for convenience and use Imandra_util.gensym to assign a name to your `src` block first, and then act by_name, so give this endpoint a try if you're having issues with /*/by-src.   # noqa: E501

        This method makes a synchronous HTTP request by default. To make an
        asynchronous HTTP request, please pass async_req=True

        >>> thread = api.instance_by_name(instance_request_name, async_req=True)
        >>> result = thread.get()

        :param instance_request_name: (required)
        :type instance_request_name: InstanceRequestName
        :param async_req: Whether to execute the request asynchronously.
        :type async_req: bool, optional
        :param _request_timeout: timeout setting for this request. If one
                                 number provided, it will be total request
                                 timeout. It can also be a pair (tuple) of
                                 (connection, read) timeouts.
        :return: Returns the result object.
                 If the method is called asynchronously,
                 returns the request thread.
        :rtype: InstanceResponse
        """
        kwargs['_return_http_data_only'] = True
        if '_preload_content' in kwargs:
            raise ValueError("Error! Please call the instance_by_name_with_http_info method with `_preload_content` instead and obtain raw data from ApiResponse.raw_data")
        return self.instance_by_name_with_http_info(instance_request_name, **kwargs)  # noqa: E501

    @validate_arguments
    def instance_by_name_with_http_info(self, instance_request_name : InstanceRequestName, **kwargs) -> ApiResponse:  # noqa: E501
        """Search for an instance from a property (boolean function) by name. (The name should exist from a previous call to `/eval`).  This is the more 'primitive' call - calls to /*/by-src are for convenience and use Imandra_util.gensym to assign a name to your `src` block first, and then act by_name, so give this endpoint a try if you're having issues with /*/by-src.   # noqa: E501

        This method makes a synchronous HTTP request by default. To make an
        asynchronous HTTP request, please pass async_req=True

        >>> thread = api.instance_by_name_with_http_info(instance_request_name, async_req=True)
        >>> result = thread.get()

        :param instance_request_name: (required)
        :type instance_request_name: InstanceRequestName
        :param async_req: Whether to execute the request asynchronously.
        :type async_req: bool, optional
        :param _preload_content: if False, the ApiResponse.data will
                                 be set to none and raw_data will store the 
                                 HTTP response body without reading/decoding.
                                 Default is True.
        :type _preload_content: bool, optional
        :param _return_http_data_only: response data instead of ApiResponse
                                       object with status code, headers, etc
        :type _return_http_data_only: bool, optional
        :param _request_timeout: timeout setting for this request. If one
                                 number provided, it will be total request
                                 timeout. It can also be a pair (tuple) of
                                 (connection, read) timeouts.
        :param _request_auth: set to override the auth_settings for an a single
                              request; this effectively ignores the authentication
                              in the spec for a single request.
        :type _request_auth: dict, optional
        :type _content_type: string, optional: force content-type for the request
        :return: Returns the result object.
                 If the method is called asynchronously,
                 returns the request thread.
        :rtype: tuple(InstanceResponse, status_code(int), headers(HTTPHeaderDict))
        """

        _params = locals()

        _all_params = [
            'instance_request_name'
        ]
        _all_params.extend(
            [
                'async_req',
                '_return_http_data_only',
                '_preload_content',
                '_request_timeout',
                '_request_auth',
                '_content_type',
                '_headers'
            ]
        )

        # validate the arguments
        for _key, _val in _params['kwargs'].items():
            if _key not in _all_params:
                raise ApiTypeError(
                    "Got an unexpected keyword argument '%s'"
                    " to method instance_by_name" % _key
                )
            _params[_key] = _val
        del _params['kwargs']

        _collection_formats = {}

        # process the path parameters
        _path_params = {}

        # process the query parameters
        _query_params = []
        # process the header parameters
        _header_params = dict(_params.get('_headers', {}))
        # process the form parameters
        _form_params = []
        _files = {}
        # process the body parameter
        _body_params = None
        if _params['instance_request_name'] is not None:
            _body_params = _params['instance_request_name']

        # set the HTTP header `Accept`
        _header_params['Accept'] = self.api_client.select_header_accept(
            ['application/json'])  # noqa: E501

        # set the HTTP header `Content-Type`
        _content_types_list = _params.get('_content_type',
            self.api_client.select_header_content_type(
                ['application/json']))
        if _content_types_list:
                _header_params['Content-Type'] = _content_types_list

        # authentication setting
        _auth_settings = ['bearerAuth']  # noqa: E501

        _response_types_map = {
            '200': "InstanceResponse",
            '422': "ResponseError",
        }

        return self.api_client.call_api(
            '/instance/by-name', 'POST',
            _path_params,
            _query_params,
            _header_params,
            body=_body_params,
            post_params=_form_params,
            files=_files,
            response_types_map=_response_types_map,
            auth_settings=_auth_settings,
            async_req=_params.get('async_req'),
            _return_http_data_only=_params.get('_return_http_data_only'),  # noqa: E501
            _preload_content=_params.get('_preload_content', True),
            _request_timeout=_params.get('_request_timeout'),
            collection_formats=_collection_formats,
            _request_auth=_params.get('_request_auth'))

    @validate_arguments
    def instance_by_src(self, instance_request_src : InstanceRequestSrc, **kwargs) -> InstanceResponse:  # noqa: E501
        """Obtain an instance from a property (boolean function) from lambda expression source  # noqa: E501

        This method makes a synchronous HTTP request by default. To make an
        asynchronous HTTP request, please pass async_req=True

        >>> thread = api.instance_by_src(instance_request_src, async_req=True)
        >>> result = thread.get()

        :param instance_request_src: (required)
        :type instance_request_src: InstanceRequestSrc
        :param async_req: Whether to execute the request asynchronously.
        :type async_req: bool, optional
        :param _request_timeout: timeout setting for this request. If one
                                 number provided, it will be total request
                                 timeout. It can also be a pair (tuple) of
                                 (connection, read) timeouts.
        :return: Returns the result object.
                 If the method is called asynchronously,
                 returns the request thread.
        :rtype: InstanceResponse
        """
        kwargs['_return_http_data_only'] = True
        if '_preload_content' in kwargs:
            raise ValueError("Error! Please call the instance_by_src_with_http_info method with `_preload_content` instead and obtain raw data from ApiResponse.raw_data")
        return self.instance_by_src_with_http_info(instance_request_src, **kwargs)  # noqa: E501

    @validate_arguments
    def instance_by_src_with_http_info(self, instance_request_src : InstanceRequestSrc, **kwargs) -> ApiResponse:  # noqa: E501
        """Obtain an instance from a property (boolean function) from lambda expression source  # noqa: E501

        This method makes a synchronous HTTP request by default. To make an
        asynchronous HTTP request, please pass async_req=True

        >>> thread = api.instance_by_src_with_http_info(instance_request_src, async_req=True)
        >>> result = thread.get()

        :param instance_request_src: (required)
        :type instance_request_src: InstanceRequestSrc
        :param async_req: Whether to execute the request asynchronously.
        :type async_req: bool, optional
        :param _preload_content: if False, the ApiResponse.data will
                                 be set to none and raw_data will store the 
                                 HTTP response body without reading/decoding.
                                 Default is True.
        :type _preload_content: bool, optional
        :param _return_http_data_only: response data instead of ApiResponse
                                       object with status code, headers, etc
        :type _return_http_data_only: bool, optional
        :param _request_timeout: timeout setting for this request. If one
                                 number provided, it will be total request
                                 timeout. It can also be a pair (tuple) of
                                 (connection, read) timeouts.
        :param _request_auth: set to override the auth_settings for an a single
                              request; this effectively ignores the authentication
                              in the spec for a single request.
        :type _request_auth: dict, optional
        :type _content_type: string, optional: force content-type for the request
        :return: Returns the result object.
                 If the method is called asynchronously,
                 returns the request thread.
        :rtype: tuple(InstanceResponse, status_code(int), headers(HTTPHeaderDict))
        """

        _params = locals()

        _all_params = [
            'instance_request_src'
        ]
        _all_params.extend(
            [
                'async_req',
                '_return_http_data_only',
                '_preload_content',
                '_request_timeout',
                '_request_auth',
                '_content_type',
                '_headers'
            ]
        )

        # validate the arguments
        for _key, _val in _params['kwargs'].items():
            if _key not in _all_params:
                raise ApiTypeError(
                    "Got an unexpected keyword argument '%s'"
                    " to method instance_by_src" % _key
                )
            _params[_key] = _val
        del _params['kwargs']

        _collection_formats = {}

        # process the path parameters
        _path_params = {}

        # process the query parameters
        _query_params = []
        # process the header parameters
        _header_params = dict(_params.get('_headers', {}))
        # process the form parameters
        _form_params = []
        _files = {}
        # process the body parameter
        _body_params = None
        if _params['instance_request_src'] is not None:
            _body_params = _params['instance_request_src']

        # set the HTTP header `Accept`
        _header_params['Accept'] = self.api_client.select_header_accept(
            ['application/json'])  # noqa: E501

        # set the HTTP header `Content-Type`
        _content_types_list = _params.get('_content_type',
            self.api_client.select_header_content_type(
                ['application/json']))
        if _content_types_list:
                _header_params['Content-Type'] = _content_types_list

        # authentication setting
        _auth_settings = ['bearerAuth']  # noqa: E501

        _response_types_map = {
            '200': "InstanceResponse",
            '422': "ResponseError",
        }

        return self.api_client.call_api(
            '/instance/by-src', 'POST',
            _path_params,
            _query_params,
            _header_params,
            body=_body_params,
            post_params=_form_params,
            files=_files,
            response_types_map=_response_types_map,
            auth_settings=_auth_settings,
            async_req=_params.get('async_req'),
            _return_http_data_only=_params.get('_return_http_data_only'),  # noqa: E501
            _preload_content=_params.get('_preload_content', True),
            _request_timeout=_params.get('_request_timeout'),
            collection_formats=_collection_formats,
            _request_auth=_params.get('_request_auth'))

    @validate_arguments
    def reset(self, **kwargs) -> str:  # noqa: E501
        """Reset Imandra internal state  # noqa: E501

        Equivalent to #reset directive in imandra-repl  # noqa: E501
        This method makes a synchronous HTTP request by default. To make an
        asynchronous HTTP request, please pass async_req=True

        >>> thread = api.reset(async_req=True)
        >>> result = thread.get()

        :param async_req: Whether to execute the request asynchronously.
        :type async_req: bool, optional
        :param _request_timeout: timeout setting for this request. If one
                                 number provided, it will be total request
                                 timeout. It can also be a pair (tuple) of
                                 (connection, read) timeouts.
        :return: Returns the result object.
                 If the method is called asynchronously,
                 returns the request thread.
        :rtype: str
        """
        kwargs['_return_http_data_only'] = True
        if '_preload_content' in kwargs:
            raise ValueError("Error! Please call the reset_with_http_info method with `_preload_content` instead and obtain raw data from ApiResponse.raw_data")
        return self.reset_with_http_info(**kwargs)  # noqa: E501

    @validate_arguments
    def reset_with_http_info(self, **kwargs) -> ApiResponse:  # noqa: E501
        """Reset Imandra internal state  # noqa: E501

        Equivalent to #reset directive in imandra-repl  # noqa: E501
        This method makes a synchronous HTTP request by default. To make an
        asynchronous HTTP request, please pass async_req=True

        >>> thread = api.reset_with_http_info(async_req=True)
        >>> result = thread.get()

        :param async_req: Whether to execute the request asynchronously.
        :type async_req: bool, optional
        :param _preload_content: if False, the ApiResponse.data will
                                 be set to none and raw_data will store the 
                                 HTTP response body without reading/decoding.
                                 Default is True.
        :type _preload_content: bool, optional
        :param _return_http_data_only: response data instead of ApiResponse
                                       object with status code, headers, etc
        :type _return_http_data_only: bool, optional
        :param _request_timeout: timeout setting for this request. If one
                                 number provided, it will be total request
                                 timeout. It can also be a pair (tuple) of
                                 (connection, read) timeouts.
        :param _request_auth: set to override the auth_settings for an a single
                              request; this effectively ignores the authentication
                              in the spec for a single request.
        :type _request_auth: dict, optional
        :type _content_type: string, optional: force content-type for the request
        :return: Returns the result object.
                 If the method is called asynchronously,
                 returns the request thread.
        :rtype: tuple(str, status_code(int), headers(HTTPHeaderDict))
        """

        _params = locals()

        _all_params = [
        ]
        _all_params.extend(
            [
                'async_req',
                '_return_http_data_only',
                '_preload_content',
                '_request_timeout',
                '_request_auth',
                '_content_type',
                '_headers'
            ]
        )

        # validate the arguments
        for _key, _val in _params['kwargs'].items():
            if _key not in _all_params:
                raise ApiTypeError(
                    "Got an unexpected keyword argument '%s'"
                    " to method reset" % _key
                )
            _params[_key] = _val
        del _params['kwargs']

        _collection_formats = {}

        # process the path parameters
        _path_params = {}

        # process the query parameters
        _query_params = []
        # process the header parameters
        _header_params = dict(_params.get('_headers', {}))
        # process the form parameters
        _form_params = []
        _files = {}
        # process the body parameter
        _body_params = None
        # set the HTTP header `Accept`
        _header_params['Accept'] = self.api_client.select_header_accept(
            ['text/plain'])  # noqa: E501

        # authentication setting
        _auth_settings = ['bearerAuth']  # noqa: E501

        _response_types_map = {
            '200': "str",
        }

        return self.api_client.call_api(
            '/reset', 'POST',
            _path_params,
            _query_params,
            _header_params,
            body=_body_params,
            post_params=_form_params,
            files=_files,
            response_types_map=_response_types_map,
            auth_settings=_auth_settings,
            async_req=_params.get('async_req'),
            _return_http_data_only=_params.get('_return_http_data_only'),  # noqa: E501
            _preload_content=_params.get('_preload_content', True),
            _request_timeout=_params.get('_request_timeout'),
            collection_formats=_collection_formats,
            _request_auth=_params.get('_request_auth'))

    @validate_arguments
    def shutdown(self, **kwargs) -> str:  # noqa: E501
        """Shutdown the server  # noqa: E501

        This method makes a synchronous HTTP request by default. To make an
        asynchronous HTTP request, please pass async_req=True

        >>> thread = api.shutdown(async_req=True)
        >>> result = thread.get()

        :param async_req: Whether to execute the request asynchronously.
        :type async_req: bool, optional
        :param _request_timeout: timeout setting for this request. If one
                                 number provided, it will be total request
                                 timeout. It can also be a pair (tuple) of
                                 (connection, read) timeouts.
        :return: Returns the result object.
                 If the method is called asynchronously,
                 returns the request thread.
        :rtype: str
        """
        kwargs['_return_http_data_only'] = True
        if '_preload_content' in kwargs:
            raise ValueError("Error! Please call the shutdown_with_http_info method with `_preload_content` instead and obtain raw data from ApiResponse.raw_data")
        return self.shutdown_with_http_info(**kwargs)  # noqa: E501

    @validate_arguments
    def shutdown_with_http_info(self, **kwargs) -> ApiResponse:  # noqa: E501
        """Shutdown the server  # noqa: E501

        This method makes a synchronous HTTP request by default. To make an
        asynchronous HTTP request, please pass async_req=True

        >>> thread = api.shutdown_with_http_info(async_req=True)
        >>> result = thread.get()

        :param async_req: Whether to execute the request asynchronously.
        :type async_req: bool, optional
        :param _preload_content: if False, the ApiResponse.data will
                                 be set to none and raw_data will store the 
                                 HTTP response body without reading/decoding.
                                 Default is True.
        :type _preload_content: bool, optional
        :param _return_http_data_only: response data instead of ApiResponse
                                       object with status code, headers, etc
        :type _return_http_data_only: bool, optional
        :param _request_timeout: timeout setting for this request. If one
                                 number provided, it will be total request
                                 timeout. It can also be a pair (tuple) of
                                 (connection, read) timeouts.
        :param _request_auth: set to override the auth_settings for an a single
                              request; this effectively ignores the authentication
                              in the spec for a single request.
        :type _request_auth: dict, optional
        :type _content_type: string, optional: force content-type for the request
        :return: Returns the result object.
                 If the method is called asynchronously,
                 returns the request thread.
        :rtype: tuple(str, status_code(int), headers(HTTPHeaderDict))
        """

        _params = locals()

        _all_params = [
        ]
        _all_params.extend(
            [
                'async_req',
                '_return_http_data_only',
                '_preload_content',
                '_request_timeout',
                '_request_auth',
                '_content_type',
                '_headers'
            ]
        )

        # validate the arguments
        for _key, _val in _params['kwargs'].items():
            if _key not in _all_params:
                raise ApiTypeError(
                    "Got an unexpected keyword argument '%s'"
                    " to method shutdown" % _key
                )
            _params[_key] = _val
        del _params['kwargs']

        _collection_formats = {}

        # process the path parameters
        _path_params = {}

        # process the query parameters
        _query_params = []
        # process the header parameters
        _header_params = dict(_params.get('_headers', {}))
        # process the form parameters
        _form_params = []
        _files = {}
        # process the body parameter
        _body_params = None
        # set the HTTP header `Accept`
        _header_params['Accept'] = self.api_client.select_header_accept(
            ['text/plain'])  # noqa: E501

        # authentication setting
        _auth_settings = ['bearerAuth']  # noqa: E501

        _response_types_map = {
            '200': "str",
        }

        return self.api_client.call_api(
            '/shutdown', 'POST',
            _path_params,
            _query_params,
            _header_params,
            body=_body_params,
            post_params=_form_params,
            files=_files,
            response_types_map=_response_types_map,
            auth_settings=_auth_settings,
            async_req=_params.get('async_req'),
            _return_http_data_only=_params.get('_return_http_data_only'),  # noqa: E501
            _preload_content=_params.get('_preload_content', True),
            _request_timeout=_params.get('_request_timeout'),
            collection_formats=_collection_formats,
            _request_auth=_params.get('_request_auth'))

    @validate_arguments
    def verify_by_name(self, verify_request_name : VerifyRequestName, **kwargs) -> VerifyResponse:  # noqa: E501
        """Verify a property (boolean function) by name. (The name should exist from a previous call to `/eval`).  This is the more 'primitive' call - calls to /*/by-src are for convenience and use Imandra_util.gensym to assign a name to your `src` block first, and then act by_name, so give this endpoint a try if you're having issues with /*/by-src.   # noqa: E501

        This method makes a synchronous HTTP request by default. To make an
        asynchronous HTTP request, please pass async_req=True

        >>> thread = api.verify_by_name(verify_request_name, async_req=True)
        >>> result = thread.get()

        :param verify_request_name: (required)
        :type verify_request_name: VerifyRequestName
        :param async_req: Whether to execute the request asynchronously.
        :type async_req: bool, optional
        :param _request_timeout: timeout setting for this request. If one
                                 number provided, it will be total request
                                 timeout. It can also be a pair (tuple) of
                                 (connection, read) timeouts.
        :return: Returns the result object.
                 If the method is called asynchronously,
                 returns the request thread.
        :rtype: VerifyResponse
        """
        kwargs['_return_http_data_only'] = True
        if '_preload_content' in kwargs:
            raise ValueError("Error! Please call the verify_by_name_with_http_info method with `_preload_content` instead and obtain raw data from ApiResponse.raw_data")
        return self.verify_by_name_with_http_info(verify_request_name, **kwargs)  # noqa: E501

    @validate_arguments
    def verify_by_name_with_http_info(self, verify_request_name : VerifyRequestName, **kwargs) -> ApiResponse:  # noqa: E501
        """Verify a property (boolean function) by name. (The name should exist from a previous call to `/eval`).  This is the more 'primitive' call - calls to /*/by-src are for convenience and use Imandra_util.gensym to assign a name to your `src` block first, and then act by_name, so give this endpoint a try if you're having issues with /*/by-src.   # noqa: E501

        This method makes a synchronous HTTP request by default. To make an
        asynchronous HTTP request, please pass async_req=True

        >>> thread = api.verify_by_name_with_http_info(verify_request_name, async_req=True)
        >>> result = thread.get()

        :param verify_request_name: (required)
        :type verify_request_name: VerifyRequestName
        :param async_req: Whether to execute the request asynchronously.
        :type async_req: bool, optional
        :param _preload_content: if False, the ApiResponse.data will
                                 be set to none and raw_data will store the 
                                 HTTP response body without reading/decoding.
                                 Default is True.
        :type _preload_content: bool, optional
        :param _return_http_data_only: response data instead of ApiResponse
                                       object with status code, headers, etc
        :type _return_http_data_only: bool, optional
        :param _request_timeout: timeout setting for this request. If one
                                 number provided, it will be total request
                                 timeout. It can also be a pair (tuple) of
                                 (connection, read) timeouts.
        :param _request_auth: set to override the auth_settings for an a single
                              request; this effectively ignores the authentication
                              in the spec for a single request.
        :type _request_auth: dict, optional
        :type _content_type: string, optional: force content-type for the request
        :return: Returns the result object.
                 If the method is called asynchronously,
                 returns the request thread.
        :rtype: tuple(VerifyResponse, status_code(int), headers(HTTPHeaderDict))
        """

        _params = locals()

        _all_params = [
            'verify_request_name'
        ]
        _all_params.extend(
            [
                'async_req',
                '_return_http_data_only',
                '_preload_content',
                '_request_timeout',
                '_request_auth',
                '_content_type',
                '_headers'
            ]
        )

        # validate the arguments
        for _key, _val in _params['kwargs'].items():
            if _key not in _all_params:
                raise ApiTypeError(
                    "Got an unexpected keyword argument '%s'"
                    " to method verify_by_name" % _key
                )
            _params[_key] = _val
        del _params['kwargs']

        _collection_formats = {}

        # process the path parameters
        _path_params = {}

        # process the query parameters
        _query_params = []
        # process the header parameters
        _header_params = dict(_params.get('_headers', {}))
        # process the form parameters
        _form_params = []
        _files = {}
        # process the body parameter
        _body_params = None
        if _params['verify_request_name'] is not None:
            _body_params = _params['verify_request_name']

        # set the HTTP header `Accept`
        _header_params['Accept'] = self.api_client.select_header_accept(
            ['application/json'])  # noqa: E501

        # set the HTTP header `Content-Type`
        _content_types_list = _params.get('_content_type',
            self.api_client.select_header_content_type(
                ['application/json']))
        if _content_types_list:
                _header_params['Content-Type'] = _content_types_list

        # authentication setting
        _auth_settings = ['bearerAuth']  # noqa: E501

        _response_types_map = {
            '200': "VerifyResponse",
            '422': "ResponseError",
        }

        return self.api_client.call_api(
            '/verify/by-name', 'POST',
            _path_params,
            _query_params,
            _header_params,
            body=_body_params,
            post_params=_form_params,
            files=_files,
            response_types_map=_response_types_map,
            auth_settings=_auth_settings,
            async_req=_params.get('async_req'),
            _return_http_data_only=_params.get('_return_http_data_only'),  # noqa: E501
            _preload_content=_params.get('_preload_content', True),
            _request_timeout=_params.get('_request_timeout'),
            collection_formats=_collection_formats,
            _request_auth=_params.get('_request_auth'))

    @validate_arguments
    def verify_by_src(self, verify_request_src : VerifyRequestSrc, **kwargs) -> VerifyResponse:  # noqa: E501
        """Verify a property (boolean function) from lambda expression source  # noqa: E501

        This method makes a synchronous HTTP request by default. To make an
        asynchronous HTTP request, please pass async_req=True

        >>> thread = api.verify_by_src(verify_request_src, async_req=True)
        >>> result = thread.get()

        :param verify_request_src: (required)
        :type verify_request_src: VerifyRequestSrc
        :param async_req: Whether to execute the request asynchronously.
        :type async_req: bool, optional
        :param _request_timeout: timeout setting for this request. If one
                                 number provided, it will be total request
                                 timeout. It can also be a pair (tuple) of
                                 (connection, read) timeouts.
        :return: Returns the result object.
                 If the method is called asynchronously,
                 returns the request thread.
        :rtype: VerifyResponse
        """
        kwargs['_return_http_data_only'] = True
        if '_preload_content' in kwargs:
            raise ValueError("Error! Please call the verify_by_src_with_http_info method with `_preload_content` instead and obtain raw data from ApiResponse.raw_data")
        return self.verify_by_src_with_http_info(verify_request_src, **kwargs)  # noqa: E501

    @validate_arguments
    def verify_by_src_with_http_info(self, verify_request_src : VerifyRequestSrc, **kwargs) -> ApiResponse:  # noqa: E501
        """Verify a property (boolean function) from lambda expression source  # noqa: E501

        This method makes a synchronous HTTP request by default. To make an
        asynchronous HTTP request, please pass async_req=True

        >>> thread = api.verify_by_src_with_http_info(verify_request_src, async_req=True)
        >>> result = thread.get()

        :param verify_request_src: (required)
        :type verify_request_src: VerifyRequestSrc
        :param async_req: Whether to execute the request asynchronously.
        :type async_req: bool, optional
        :param _preload_content: if False, the ApiResponse.data will
                                 be set to none and raw_data will store the 
                                 HTTP response body without reading/decoding.
                                 Default is True.
        :type _preload_content: bool, optional
        :param _return_http_data_only: response data instead of ApiResponse
                                       object with status code, headers, etc
        :type _return_http_data_only: bool, optional
        :param _request_timeout: timeout setting for this request. If one
                                 number provided, it will be total request
                                 timeout. It can also be a pair (tuple) of
                                 (connection, read) timeouts.
        :param _request_auth: set to override the auth_settings for an a single
                              request; this effectively ignores the authentication
                              in the spec for a single request.
        :type _request_auth: dict, optional
        :type _content_type: string, optional: force content-type for the request
        :return: Returns the result object.
                 If the method is called asynchronously,
                 returns the request thread.
        :rtype: tuple(VerifyResponse, status_code(int), headers(HTTPHeaderDict))
        """

        _params = locals()

        _all_params = [
            'verify_request_src'
        ]
        _all_params.extend(
            [
                'async_req',
                '_return_http_data_only',
                '_preload_content',
                '_request_timeout',
                '_request_auth',
                '_content_type',
                '_headers'
            ]
        )

        # validate the arguments
        for _key, _val in _params['kwargs'].items():
            if _key not in _all_params:
                raise ApiTypeError(
                    "Got an unexpected keyword argument '%s'"
                    " to method verify_by_src" % _key
                )
            _params[_key] = _val
        del _params['kwargs']

        _collection_formats = {}

        # process the path parameters
        _path_params = {}

        # process the query parameters
        _query_params = []
        # process the header parameters
        _header_params = dict(_params.get('_headers', {}))
        # process the form parameters
        _form_params = []
        _files = {}
        # process the body parameter
        _body_params = None
        if _params['verify_request_src'] is not None:
            _body_params = _params['verify_request_src']

        # set the HTTP header `Accept`
        _header_params['Accept'] = self.api_client.select_header_accept(
            ['application/json'])  # noqa: E501

        # set the HTTP header `Content-Type`
        _content_types_list = _params.get('_content_type',
            self.api_client.select_header_content_type(
                ['application/json']))
        if _content_types_list:
                _header_params['Content-Type'] = _content_types_list

        # authentication setting
        _auth_settings = ['bearerAuth']  # noqa: E501

        _response_types_map = {
            '200': "VerifyResponse",
            '422': "ResponseError",
        }

        return self.api_client.call_api(
            '/verify/by-src', 'POST',
            _path_params,
            _query_params,
            _header_params,
            body=_body_params,
            post_params=_form_params,
            files=_files,
            response_types_map=_response_types_map,
            auth_settings=_auth_settings,
            async_req=_params.get('async_req'),
            _return_http_data_only=_params.get('_return_http_data_only'),  # noqa: E501
            _preload_content=_params.get('_preload_content', True),
            _request_timeout=_params.get('_request_timeout'),
            collection_formats=_collection_formats,
            _request_auth=_params.get('_request_auth'))



"""
Imandra Universe Cogito Agent.

Example:
    $ pip install 'imandra[universe]'
    $ export IMANDRA_API_KEY=<redacted>
    $ python
    >>> from imandra.u.agents import cogito
    >>> from langchain_core.messages import HumanMessage
    >>> g = cogito.get_remote_graph()
    >>> cogito.create_thread_sync(g)
    >>> response = g.invoke({'messages': [HumanMessage(content='hello')]})
    >>> response['messages'][2]['content']
   'Hello! How can I assist you today?'
"""

import inspect
from typing import Optional

from ... import auth

try:
    from langgraph.pregel.remote import (  # pyright: ignore [reportMissingImports]
        RemoteGraph,
        RunnableConfig,
    )
except ModuleNotFoundError as err:
    note = """
        Install imandra with the optional 'universe' dependency to enable imandra.u.agents.cogito:

            pip install 'imandra[universe]>=2.0.0'
    """
    err.msg += "\n\n" + inspect.cleandoc(note)
    raise


def get_remote_graph(
    config: Optional[RunnableConfig] = None,
    api_key=None,
    scheme=None,
    host=None,
    api_version=None,
) -> RemoteGraph:
    """Create a RemoteGraph configured for Cogito via Imandra Universe."""
    c = auth.Config(api_key=api_key, scheme=scheme, host=host, api_version=api_version)
    url = f"{c.get_url()}/agents/cogito"

    config = config or {}
    config.setdefault("configurable", {}).setdefault("imandra_api_key", api_key)

    remote_graph = RemoteGraph(
        "cogito",
        url=url,
        headers=c.get_headers(),
        config=config,
    )

    return remote_graph


def create_thread_sync(remote_graph: RemoteGraph):
    """Create a thread and configure the RemoteGraph to use it."""
    if remote_graph.sync_client is not None and remote_graph.config is not None:
        thread = remote_graph.sync_client.threads.create()
        remote_graph.config.setdefault("configurable", {})["thread_id"] = thread[
            "thread_id"
        ]



import urllib
import urllib.parse
import urllib.request
import urllib.error
import gzip


def synth(auth, file_path):
    with open(file_path, "rb") as f:
        file_contents = f.read()
        content = gzip.compress(file_contents)

    url = "{}/{}".format(auth.imandra_web_host, "api/imandra-rule-synth/synth")
    clen = len(content)
    headers = {
        "X-Auth": auth.token,
        "Content-Length": clen,
        "Content-Encoding": "gzip",
        "Accept-Encoding": "gzip",
    }
    request = urllib.request.Request(url, content, headers=headers)

    try:
        with urllib.request.urlopen(request) as response:
            if response.headers["content-encoding"] == "gzip":
                return gzip.decompress(response.read()).decode("utf-8")
            else:
                return response.read().decode("utf-8")

    except urllib.error.HTTPError as e:
        raise ValueError(e.read().decode("utf-8"))



from .api import auth, instance
import imandra_http_api_client

from imandra_http_api_client.models import WithUnknownReasonBody, UpToBody
from imandra_http_api_client.models import InstanceResponse, InstanceResult
from imandra_http_api_client.models import VerifyResponse, VerifyResult


def instance_monkey_str(self):
    if self.type == InstanceResult.SAT:
        src = self.body.actual_instance.instance.model.src
        return "Instance found:\n{}".format(src.strip())
    elif self.type == InstanceResult.UNSAT:
        return "Unsatisfiable"
    elif self.type == InstanceResult.UNSAT_UPTO:
        steps = self.body.actual_instance.steps
        return "Unsatisfiable upto {}".format(steps)
    elif self.type == InstanceResult.UNKNOWN:
        instance = self.body.actual_instance
        if instance.__class__ == WithUnknownReasonBody:
            reason = instance.unknown_reason
            return "Unknown: {}".format(reason.strip())
        elif instance.__class__ == UpToBody:
            steps = instance.steps
            return "Unknown: Searched Up To {}".format(steps)
        return "Unknown"


def verify_monkey_str(self):
    if self.type == VerifyResult.REFUTED:
        src = self.body.actual_instance.instance.model.src
        return "Refuted, with counterexample:\n{}".format(src.strip())
    elif self.type == VerifyResult.PROVED:
        return "Proved"
    elif self.type == VerifyResult.PROVED_UPTO:
        steps = self.body.actual_instance.steps
        return "Proved upto {}".format(steps)
    elif self.type == VerifyResult.UNKNOWN:
        instance = self.body.actual_instance
        if instance.__class__ == WithUnknownReasonBody:
            reason = instance.unknown_reason
            return "Unknown: {}".format(reason.strip())
        elif instance.__class__ == UpToBody:
            steps = instance.steps
            return "Unknown: Verified Up To {}".format(steps)
        return "Unknown"


InstanceResponse.__str__ = instance_monkey_str  # type: ignore
VerifyResponse.__str__ = verify_monkey_str  # type: ignore


class HttpInstanceSession:
    def __init__(self):
        self.auth = auth.Auth()
        self.instance_ = instance.create(self.auth, None, "imandra-http-api")

        config = imandra_http_api_client.Configuration(
            host=self.instance_["new_pod"]["url"],
            access_token=self.instance_["new_pod"]["exchange_token"],
        )

        self.api_client = imandra_http_api_client.ApiClient(config)
        self.api_instance = imandra_http_api_client.DefaultApi(self.api_client)

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()

    def close(self):
        instance.delete(self.auth, self.instance_["new_pod"]["id"])

    def get_history(self):
        return self.api_instance.get_history()

    def reset(self):
        return self.api_instance.reset()

    def eval(self, src, **args):
        eval_request_src = imandra_http_api_client.EvalRequestSrc(src=src, **args)
        try:
            return self.api_instance.eval(eval_request_src)
        except Exception as err:
            return err

    def verify(self, src, **args):
        verify_request_src = imandra_http_api_client.VerifyRequestSrc(src=src, **args)
        return self.api_instance.verify_by_src(verify_request_src)

    def instance(self, src, **args):
        instance_request_src = imandra_http_api_client.InstanceRequestSrc(
            src=src, **args
        )
        return self.api_instance.instance_by_src(instance_request_src)

    def decompose(self, name, **args):
        decompose_request = imandra_http_api_client.DecomposeRequestSrc(
            name=name, **args
        )
        return self.api_instance.decompose(decompose_request)



import urllib
import urllib.parse
import urllib.request
import urllib.error
import json


def list(auth):
    url = "{}/{}".format(auth.imandra_web_host, "api/instances/list")
    headers = {"X-Auth": auth.token}

    request = urllib.request.Request(url, headers=headers)

    try:
        response = urllib.request.urlopen(request)
    except urllib.error.HTTPError as e:
        raise ValueError(e.read().decode("utf-8"))

    instances = json.loads(response.read())
    return instances


def delete(auth, instance_id):
    url = "{}/{}/{}".format(auth.imandra_web_host, "api/instances/delete", instance_id)
    headers = {"X-Auth": auth.token}

    data = {}
    request = urllib.request.Request(url, data, headers=headers)

    try:
        _ = urllib.request.urlopen(request)
    except urllib.error.HTTPError as e:
        raise ValueError(e.read().decode("utf-8"))


def create(auth, version, instance_type):
    url = "{}/{}".format(auth.imandra_web_host, "api/instances")
    headers = {"X-Auth": auth.token, "Content-Type": "application/json"}

    req = {}
    req["cluster"] = auth.zone
    req["instance_type"] = instance_type
    if version:
        req["version"] = version

    data = json.dumps(req).encode("utf-8")

    request = urllib.request.Request(url, data, headers=headers)

    try:
        response = urllib.request.urlopen(request)
    except urllib.error.HTTPError as e:
        raise ValueError(e.read().decode("utf-8"))

    s = response.read()
    instance_response = json.loads(s)

    return instance_response



import json
import os
import shutil
import urllib.request
import tempfile
from sys import platform

from .. import api
from ..api.auth import Auth
from .ipl_log_analysis import IplLogAnalysis


def install_imandra_core():
    if os.name == "nt":
        print(
            "Imandra Core cannot be installed on native Windows.\n\nPlease install the Imandra CLI on Windows Subsystem for Linux (WSL), and re-run this command there."
        )
        exit(1)

    with urllib.request.urlopen(
        "https://storage.googleapis.com/imandra-installer/install.sh"
    ) as fi:
        (fd, path) = tempfile.mkstemp()
        with os.fdopen(fd, "w") as fo:
            fo.write(fi.read().decode("utf-8"))

    os.execvp("sh", ["sh"] + [path])


def run_repl(auth, args):
    if shutil.which(auth.imandra_repl):
        os.execvp(auth.imandra_repl, [auth.imandra_repl] + args)
    else:
        print(
            "imandra-repl is not installed. Run: 'imandra core install' to install it."
        )


def run_ipl_log_analysis(
    auth,
    file,
    traces,
    organization,
    callback,
    interactive,
    imandra_host,
    imandra_port,
    mode,
    uuid,
    json_out,
    decomp_job_id,
    runner_image,
    skip_cache,
    sender_comp_id,
):
    imandra_port = int(imandra_port) if imandra_port is not None else None

    analysis = IplLogAnalysis(
        auth,
        file,
        traces,
        organization,
        callback,
        interactive,
        imandra_host,
        imandra_port,
        json_out,
        decomp_job_id,
        runner_image,
        sender_comp_id,
    )

    if mode == "docker":
        docker_mode = True
    elif mode == "binary":
        if platform.startswith("linux"):
            docker_mode = False
        else:
            print("Binary mode is only available on Linux.")
            return
    else:
        print(f"Unknown mode '{mode}'.")
        return

    if file is not None and not os.path.isfile(file):
        print(f"not found: {file}")
        return
    if not os.path.isdir(traces):
        print(f"not found: {traces}")
        return

    analysis.login()

    local_uuid = analysis.find_local_job() if not skip_cache else None
    if local_uuid:
        print(f"Using existing job with UUID {local_uuid} for this IPL model.")
        print(
            "Run this command with --skip-cache to force the submission of a new job.\n"
        )
        analysis.use_job(local_uuid)

        if file is not None:
            print("WARNING: ignoring --file parameter\n")
        if decomp_job_id is not None:
            print("WARNING: ignoring --decomp_job_id parameter\n")
    elif uuid is not None:
        print(f"Using provided job with UUID {uuid}")
        analysis.use_job(uuid)

        if file is not None:
            print("WARNING: ignoring --file parameter")
        if decomp_job_id is not None:
            print("WARNING: ignoring --decomp_job_id parameter")
        analysis.prepare_job_data()
    else:
        job_id = analysis.submit_job()
        print("Processing new job with UUID {}...".format(job_id))

        job_res = api.ipl.wait(auth, job_id, interval=5)
        if job_res == "done":
            print("Job ready.")
        else:
            print("Job failed.")
            return
        analysis.prepare_job_data()

    print("Starting Imandra HTTP API...")
    imandra_ready = analysis.start_imandra(10)
    if not imandra_ready:
        print("Failed to start Imandra HTTP API (timed out after 10 seconds).")
        return

    if docker_mode:
        analysis.run_docker()
    else:
        analysis.run_binary()


def main():
    import sys
    import argparse

    parser = argparse.ArgumentParser(prog="imandra", description="Imandra CLI")

    parser.set_defaults(run=lambda _: parser.print_help())
    subparsers = parser.add_subparsers()

    auth = Auth()

    parser_auth = subparsers.add_parser("auth")
    parser_auth.set_defaults(run=lambda _: parser_auth.print_help())
    parser_auth_subparsers = parser_auth.add_subparsers()

    def auth_login(_):
        auth.login()
        print("Logged in")

    parser_login = parser_auth_subparsers.add_parser("login")
    parser_login.set_defaults(run=auth_login)

    def auth_logout(_):
        auth.logout()
        print("Logged out")

    parser_logout = parser_auth_subparsers.add_parser("logout")
    parser_logout.set_defaults(run=auth_logout)

    def auth_export(_):
        print(json.dumps(auth.export()))

    parser_export = parser_auth_subparsers.add_parser("export")
    parser_export.set_defaults(run=auth_export)

    def auth_import(_):
        lines = []
        for line in sys.stdin:
            lines.append(line)

        details = json.loads("\n".join(lines))
        auth.import_(details)

    parser_import = parser_auth_subparsers.add_parser("import")
    parser_import.set_defaults(run=auth_import)

    parser_ipl = subparsers.add_parser("ipl")
    parser_ipl.set_defaults(run=lambda _: parser_ipl.print_help())
    parser_ipl_subparsers = parser_ipl.add_subparsers()

    def ipl_decompose(args):
        if args.file is not None and args.parent_job_id is not None:
            print(
                "error: --file and --parent-job-id are mutually exclusive operations. "
                "You must pick one or the other"
            )
            sys.exit(1)
        if args.file is None and args.parent_job_id is None:
            print("error: --file must be provided")
        else:
            auth.login()
            job_id = api.ipl.decompose(
                auth,
                args.file,
                args.testgen_lang,
                args.organization,
                args.callback,
                args.doc_gen,
                args.parent_job_id,
            )
            print("{}/ipl/jobs/{}".format(auth.imandra_web_host, job_id))
            if args.wait:
                status = api.ipl.wait(auth, job_id)
                print(status)

    parser_ipl_decompose = parser_ipl_subparsers.add_parser("decompose")

    parser_ipl_decompose.add_argument("--file")
    parser_ipl_decompose.add_argument("--callback")
    parser_ipl_decompose.add_argument("--testgen-lang", help=argparse.SUPPRESS)
    parser_ipl_decompose.add_argument("--organization")
    parser_ipl_decompose.add_argument(
        "--doc-gen", choices=["true", "false"], default="true"
    )
    parser_ipl_decompose.add_argument(
        "--parent-job-id",
        help=(
            "The id of a previously completed job. Instead of running a "
            "full decomposition, re-run just the script generation phase of "
            "this parent-job. When using this option all other options will be "
            "ignore and the original values from the parent-job will be used."
        ),
    )
    parser_ipl_decompose.add_argument(
        "--wait",
        action="store_true",
        help=("Block until the decomp job has finished running"),
    )
    parser_ipl_decompose.set_defaults(run=ipl_decompose)

    def ipl_unsat_analysis(args):
        if args.file is None:
            print("error: --file must be provided")
        else:
            auth.login()
            job_id = api.ipl.unsat_analysis(
                auth, args.file, args.organization, args.callback
            )
            print("{}".format(job_id))

            if args.wait:
                status = api.ipl.wait(auth, job_id)
                print(status)

    parser_ipl_unsat_analysis = parser_ipl_subparsers.add_parser("unsat-analysis")

    parser_ipl_unsat_analysis.add_argument("--file")
    parser_ipl_unsat_analysis.add_argument("--callback")
    parser_ipl_unsat_analysis.add_argument("--organization")
    parser_ipl_unsat_analysis.add_argument(
        "--wait",
        action="store_true",
        help=("Block until the unsat analysis job has finished running"),
    )
    parser_ipl_unsat_analysis.set_defaults(run=ipl_unsat_analysis)

    def ipl_log_analysis(args):
        if args.file is None and args.uuid is None:
            print("error: either --file or --uuid must be provided")
        elif args.traces is None:
            print("error: --traces must be provided")
        else:
            run_ipl_log_analysis(
                auth,
                args.file,
                args.traces,
                args.organization,
                args.callback,
                args.interactive,
                args.imandra_host,
                args.imandra_port,
                args.mode,
                args.uuid,
                args.json_out,
                args.decomp_job_id,
                args.runner_image,
                args.skip_cache,
                args.sender_comp_id,
            )

    parser_ipl_log_analysis = parser_ipl_subparsers.add_parser("log-analysis")
    parser_ipl_log_analysis.add_argument("--file")
    parser_ipl_log_analysis.add_argument("--traces")
    parser_ipl_log_analysis.add_argument("--callback")
    parser_ipl_log_analysis.add_argument("--organization")
    parser_ipl_log_analysis.add_argument("--imandra_host")
    parser_ipl_log_analysis.add_argument("--imandra_port")
    parser_ipl_log_analysis.add_argument("--uuid")
    parser_ipl_log_analysis.add_argument("--json-out")
    parser_ipl_log_analysis.add_argument("--decomp-job-id")
    parser_ipl_log_analysis.add_argument("--runner-image")
    parser_ipl_log_analysis.add_argument("--sender-comp-id")
    parser_ipl_log_analysis.add_argument(
        "--interactive", choices=["true", "false"], default="false"
    )
    parser_ipl_log_analysis.add_argument(
        "--mode", choices=["docker", "binary"], default="docker"
    )
    parser_ipl_log_analysis.add_argument("--skip-cache", action="store_true")
    parser_ipl_log_analysis.set_defaults(run=ipl_log_analysis)

    def ipl_simulator(args):
        if args.file is None:
            print("error: --file must be provided")
        else:
            auth.login()
            resp = api.ipl.simulator(auth, args.file)
            urls = dict(resp["new_pod"]["urls"])
            print("Simulator available at: {}/".format(urls["http"]))

    parser_ipl_simulator = parser_ipl_subparsers.add_parser("simulator")
    parser_ipl_simulator.add_argument("--file")
    parser_ipl_simulator.set_defaults(run=ipl_simulator)

    def ipl_status(args):
        if args.uuid is None:
            print("error: --uuid must be provided")
        else:
            auth.login()
            status = api.ipl.status(auth, args.uuid)
            print(status)

    parser_ipl_status = parser_ipl_subparsers.add_parser("status")
    parser_ipl_status.add_argument("--uuid")
    parser_ipl_status.set_defaults(run=ipl_status)

    def ipl_data(args):
        if args.uuid is None:
            print("error: --uuid must be provided")
        else:
            auth.login()
            response = api.ipl.data(auth, args.uuid, args.stdout)
            data = response["content"]
            if args.stdout:
                print(data.decode("utf-8"))
            else:
                filename = args.uuid + (
                    "" if response["content_type"] == "text/plain" else ".tar.gz"
                )

                with open(filename, "wb") as data_file:
                    data_file.write(data)
                print(filename)

    parser_ipl_data = parser_ipl_subparsers.add_parser("data")
    parser_ipl_data.add_argument("--uuid")
    parser_ipl_data.add_argument("--stdout", action="store_true")
    parser_ipl_data.set_defaults(run=ipl_data)

    def ipl_list_jobs(args):
        auth.login()
        job_type = "ipl-log-analysis-builder" if args.log_analysis else None
        resp = api.ipl.list_jobs(auth, limit=10, job_type=job_type)

        if len(resp["jobs"]) == 0:
            print("No jobs submitted yet.")
        else:
            colsfmt = "{:<36} {:<16} {:<20} {:<30} {:<30} {:<30}"
            print(
                colsfmt.format(
                    "ID", "Status", "Filename", "Submitted", "Started", "Ended"
                )
            )
            for job in resp["jobs"]:
                endTime = "-"
                if job["status"] == "cancelled":
                    endTime = job["cancelledAt"]
                elif job["status"] == "done":
                    endTime = job["completedAt"]
                elif job["status"] == "error":
                    endTime = job["failedAt"]

                filename = job["iplFile"]["filename"]
                if len(filename) > 20:
                    filename = filename[:17] + "..."

                print(
                    colsfmt.format(
                        job["id"],
                        job["status"],
                        filename,
                        job["queuedAt"],
                        job["startedAt"],
                        endTime,
                    )
                )

    parser_ipl_jobs = parser_ipl_subparsers.add_parser("list-jobs")
    parser_ipl_jobs.add_argument("--log-analysis", action="store_true")
    parser_ipl_jobs.set_defaults(run=ipl_list_jobs)

    def ipl_cancel(args):
        if args.uuid is None:
            print("error: --uuid must be provided")
        else:
            auth.login()
            api.ipl.cancel(auth, args.uuid)
            print("Cancel requested for job: {}".format(args.uuid))

    parser_ipl_cancel = parser_ipl_subparsers.add_parser("cancel")
    parser_ipl_cancel.add_argument("--uuid")
    parser_ipl_cancel.set_defaults(run=ipl_cancel)

    def ipl_wait(args):
        if args.uuid is None:
            print("error: --uuid must be provided")
        else:
            auth.login()
            status = api.ipl.wait(auth, args.uuid)
            print(status)

    parser_ipl_wait = parser_ipl_subparsers.add_parser("wait")
    parser_ipl_wait.add_argument("--uuid")
    parser_ipl_wait.set_defaults(run=ipl_wait)

    parser_core = subparsers.add_parser("core")
    parser_core.set_defaults(run=lambda _: parser_core.print_help())
    parser_core_subparsers = parser_core.add_subparsers()

    parser_instances = parser_core_subparsers.add_parser("instances")
    parser_instances.set_defaults(run=lambda _: parser_instances.print_help())
    parser_instances_subparsers = parser_instances.add_subparsers()

    def instance_list(_):
        auth.login()
        instances = api.instance.list(auth)
        print("Instances: \n")
        for i in instances:
            print(" [{}] [{}] {}".format(i["assigned_at"], i["pod_type"], i["pod_id"]))

    parser_instances_list = parser_instances_subparsers.add_parser("list")
    parser_instances_list.set_defaults(run=instance_list)

    def instance_create(args):
        if args.instance_type is None:
            print("error: --instance-type must be provided")
        else:
            auth.login()
            auth.ensure_zone()
            response = api.instance.create(auth, args.version, args.instance_type)
            print("Instance created:")
            print("- url: {}".format(response["new_pod"]["url"]))
            print("- token: {}".format(response["new_pod"]["exchange_token"]))

    parser_instances_create = parser_instances_subparsers.add_parser("create")
    parser_instances_create.add_argument("--instance-type")
    parser_instances_create.add_argument("--version")
    parser_instances_create.set_defaults(run=instance_create)

    def instance_kill(args):
        if args.id is None:
            print("error: --id must be provided")
        else:
            auth.login()
            api.instance.delete(auth, args.id)
            print("Instance killed")

    parser_instances_kill = parser_instances_subparsers.add_parser("kill")
    parser_instances_kill.add_argument("--id")
    parser_instances_kill.set_defaults(run=instance_kill)

    def rule_synth(args):
        if args.file is None:
            print("error: --file must be provided")
        else:
            auth.login()
            response = api.rule_synth.synth(auth, args.file)
            print(response)

    parser_rule_synth = subparsers.add_parser("rule-synth")
    parser_rule_synth.set_defaults(run=lambda _: parser_rule_synth.print_help())
    parser_rule_synth_subparsers = parser_rule_synth.add_subparsers()

    parser_synth = parser_rule_synth_subparsers.add_parser("synth")
    parser_synth.add_argument("--file")
    parser_synth.set_defaults(run=rule_synth)

    def cfb_analyze(args):
        if args.job is None:
            print("error: --job must be provided")
        elif args.path is None:
            print("error: --path must be provided")
        else:
            auth.login()
            response = api.cfb.analyze(auth, args.job, args.path)
            print(response)

    parser_cfb = subparsers.add_parser("cfb")
    parser_cfb.set_defaults(run=lambda _: parser_cfb.print_help())
    parser_cfb_subparsers = parser_cfb.add_subparsers()

    parser_cfb_analyze = parser_cfb_subparsers.add_parser("analyze")
    parser_cfb_analyze.add_argument("--job")
    parser_cfb_analyze.add_argument("--path")
    parser_cfb_analyze.set_defaults(run=cfb_analyze)

    def repl(args):
        run_repl(auth, args.args)

    parser_repl = parser_core_subparsers.add_parser("repl")
    parser_repl.add_argument("args", nargs=argparse.REMAINDER)
    parser_repl.set_defaults(run=repl)

    parser_install = parser_core_subparsers.add_parser("install")
    parser_install.set_defaults(run=install_imandra_core)

    if len(sys.argv) > 2 and sys.argv[1] == "core" and sys.argv[2] == "repl":
        # Run this directly so we can pass --help directly to the repl command to see its own help text
        run_repl(auth, sys.argv[3:])
    else:
        args = parser.parse_args()
        args.run(args)



# coding: utf-8

# flake8: noqa
"""
    imandra-http-server

    Query Imandra via HTTP. See also https://github.com/aestheticIntegration/bs-imandra-client for a sample client implementation and OCaml API types.  # noqa: E501

    The version of the OpenAPI document: 0.0.1
    Generated by OpenAPI Generator (https://openapi-generator.tech)

    Do not edit the class manually.
"""


# import models into model package
from imandra_http_api_client.models.decompose_region import DecomposeRegion
from imandra_http_api_client.models.decompose_request_src import DecomposeRequestSrc
from imandra_http_api_client.models.decompose_result import DecomposeResult
from imandra_http_api_client.models.eval_request_src import EvalRequestSrc
from imandra_http_api_client.models.eval_response import EvalResponse
from imandra_http_api_client.models.hints import Hints
from imandra_http_api_client.models.induct_functional_body import InductFunctionalBody
from imandra_http_api_client.models.induct_structural_body import InductStructuralBody
from imandra_http_api_client.models.induct_structural_style import InductStructuralStyle
from imandra_http_api_client.models.induct_type import InductType
from imandra_http_api_client.models.instance_request_name import InstanceRequestName
from imandra_http_api_client.models.instance_request_src import InstanceRequestSrc
from imandra_http_api_client.models.instance_response import InstanceResponse
from imandra_http_api_client.models.instance_result import InstanceResult
from imandra_http_api_client.models.instance_type import InstanceType
from imandra_http_api_client.models.method import Method
from imandra_http_api_client.models.method_body import MethodBody
from imandra_http_api_client.models.method_ext_solver_body import MethodExtSolverBody
from imandra_http_api_client.models.method_induct_body import MethodInductBody
from imandra_http_api_client.models.method_induct_body_body import MethodInductBodyBody
from imandra_http_api_client.models.method_type import MethodType
from imandra_http_api_client.models.method_unroll_body import MethodUnrollBody
from imandra_http_api_client.models.printer_details import PrinterDetails
from imandra_http_api_client.models.response_error import ResponseError
from imandra_http_api_client.models.response_instance import ResponseInstance
from imandra_http_api_client.models.response_model import ResponseModel
from imandra_http_api_client.models.syntax import Syntax
from imandra_http_api_client.models.up_to_body import UpToBody
from imandra_http_api_client.models.verify_request_name import VerifyRequestName
from imandra_http_api_client.models.verify_request_src import VerifyRequestSrc
from imandra_http_api_client.models.verify_response import VerifyResponse
from imandra_http_api_client.models.verify_response_body import VerifyResponseBody
from imandra_http_api_client.models.verify_result import VerifyResult
from imandra_http_api_client.models.with_instance_body import WithInstanceBody
from imandra_http_api_client.models.with_unknown_reason_body import WithUnknownReasonBody



import os
import sys
from typing import Dict, Optional


# Note: platformdirs distinguishes Darwin and Unix, but we use the Unix
# conventions on Darwin.
if sys.platform == "win32":
    from platformdirs.windows import Windows as PlatformDirs
else:
    from platformdirs.unix import Unix as PlatformDirs


class NoApiKey(Exception):
    pass


def _api_key_from_env():
    return os.environ.get("IMANDRA_API_KEY")


def _api_key_from_file(fpath):
    try:
        with open(fpath) as f:
            return f.read().strip()
    except OSError:
        return None


def _get_api_key(api_key=None) -> str:
    api_key = (
        api_key
        or _api_key_from_env()
        or _api_key_from_file(
            PlatformDirs(appname="imandra").user_config_path / "api_key"
        )
        or _api_key_from_file(
            PlatformDirs(appname="imandrax").user_config_path / "api_key"
        )
    )
    if api_key is None:
        raise NoApiKey("Please provide an API key.")
    return api_key


def _scheme_from_env():
    return os.environ.get("IMANDRA_API_SCHEME")


def _host_from_env():
    return os.environ.get("IMANDRA_API_HOST")


def _api_version_from_env():
    return os.environ.get("IMANDRA_API_VERSION")


class Config:
    def __init__(
        self,
        api_key: Optional[str] = None,
        scheme: Optional[str] = None,
        host: Optional[str] = None,
        api_version: Optional[str] = None,
    ):
        self._api_key = _get_api_key(api_key)
        self._scheme = scheme or _scheme_from_env() or "https"
        self._host = host or _host_from_env() or "api.imandra.ai"
        self._api_version = api_version or _api_version_from_env() or "v1beta1"

    def get_api_key(self) -> str:
        return self._api_key

    def get_headers(self) -> Dict[str, str]:
        return {"Authorization": f"Bearer {self._api_key}"}

    def get_url(self) -> str:
        return f"{self._scheme}://{self._host}/{self._api_version}"



import json
import os
import os.path
from typing import Any, Dict, TypedDict
import urllib
import urllib.parse
import urllib.request
import urllib.error
import time


class Config(TypedDict):
    host: str
    headers: Dict[str, Any]


def status(config: Config, path: str, job_id):
    path = path.format(job_id)
    url = "{}/{}".format(config["host"], path)

    request = urllib.request.Request(url, headers=config["headers"])

    try:
        response = urllib.request.urlopen(request)
        resp = response.read().decode("utf-8")
    except urllib.error.HTTPError as e:
        raise ValueError(e.read().decode("utf-8"))

    return json.loads(resp)["status"]


def wait(config: Config, path: str, job_id, interval):
    time.sleep(interval)
    s = status(config, path, job_id)
    if s in ["queued", "processing"]:
        return wait(config, path, job_id, interval)
    else:
        return s


def decompose(
    config: Config,
    path: str,
    file,
    testgen_lang,
    organization,
    callback,
    doc_gen,
    parent_job_id,
):
    params_dict = {"lang": "ipl"}
    if parent_job_id is not None:
        params_dict["parent-job-id"] = parent_job_id

    if file is not None:
        filename = os.path.basename(file)
        params_dict["filename"] = filename
        with open(file, "r") as ipl_file:
            content = ipl_file.read()
        data = content.encode("utf-8")
    else:
        data = b""

    if testgen_lang is not None:
        params_dict["testgen-lang"] = testgen_lang

    if doc_gen is not None:
        params_dict["doc-gen"] = doc_gen

    if organization is not None:
        params_dict["organization-id"] = organization

    if callback is not None:
        params_dict["callback"] = callback

    params = urllib.parse.urlencode(params_dict)
    url = "{}/{}?{}".format(config["host"], path, params)
    request = urllib.request.Request(url, data, headers=config["headers"])

    try:
        response = urllib.request.urlopen(request)
    except urllib.error.HTTPError as e:
        raise ValueError(e.read().decode("utf-8"))

    job_id = response.read().decode("utf-8")
    return job_id


def data(config: Config, path: str, job_id, file=None):
    params = f"file={file}" if file is not None else ""
    path = path.format(job_id)
    path = "{}?{}".format(path, params)
    url = "{}/{}".format(config["host"], path)

    request = urllib.request.Request(url, headers=config["headers"])

    try:
        response = urllib.request.urlopen(request)
        content = response.read()
        content_type = response.headers.get("Content-Type")
        return {
            "content_type": content_type,
            "content": content,
        }
    except urllib.error.HTTPError as e:
        if e.code == 302:
            content = e.read()
            content_type = e.headers.get("Content-Type")
            return {
                "content_type": content_type,
                "content": content,
            }
        else:
            raise ValueError(e.read().decode("utf-8"))


def simulator(config: Config, zone: str, path: str, file):
    with open(file, "r") as ipl_file:
        content = ipl_file.read()
    url = "{}/{}".format(config["host"], path)

    req = {"payload": content, "cluster": zone, "version": "latest"}

    data = json.dumps(req)
    clen = len(data)
    data = data.encode("utf-8")
    headers = config["headers"].copy()
    headers["Content-Type"] = "application/json"
    headers["Content-Length"] = clen

    request = urllib.request.Request(url, data, headers=headers)

    try:
        response = urllib.request.urlopen(request)
    except urllib.error.HTTPError as e:
        raise ValueError(e.read().decode("utf-8"))

    resp = json.loads(response.read())
    return resp


def list_jobs(config: Config, path: str, limit=10, job_type=None):
    path = f"{path}?limit={limit}"
    if job_type:
        path = f"{path}&job-type={job_type}"
    url = "{}/{}".format(config["host"], path)

    request = urllib.request.Request(url, headers=config["headers"])

    try:
        response = urllib.request.urlopen(request)
    except urllib.error.HTTPError as e:
        raise ValueError(e.read().decode("utf-8"))

    resp = json.loads(response.read())
    return resp


def unsat_analysis(config: Config, path: str, file, organization, callback):
    params_dict = {"lang": "ipl"}

    if file is not None:
        filename = os.path.basename(file)
        params_dict["filename"] = filename
        with open(file, "r") as ipl_file:
            content = ipl_file.read()
        data = content.encode("utf-8")
    else:
        data = b""

    if organization is not None:
        params_dict["organization-id"] = organization

    if callback is not None:
        params_dict["callback"] = callback

    params = urllib.parse.urlencode(params_dict)
    url = "{}/{}?{}".format(config["host"], path, params)
    request = urllib.request.Request(url, data, headers=config["headers"])

    try:
        response = urllib.request.urlopen(request)
    except urllib.error.HTTPError as e:
        raise ValueError(e.read().decode("utf-8"))

    job_id = response.read().decode("utf-8")
    return job_id


def log_analysis_builder(
    config: Config,
    path: str,
    file,
    organization=None,
    callback=None,
    decomp_job_id=None,
):

    filename = os.path.basename(file)
    params_dict = {"filename": filename}
    with open(file, "r") as ipl_file:
        content = ipl_file.read()
        file_contents = content.encode("utf-8")

    if organization is not None:
        params_dict["organization-id"] = organization

    if callback is not None:
        params_dict["callback"] = callback

    if decomp_job_id is not None:
        params_dict["decomp-job-id"] = decomp_job_id

    params = urllib.parse.urlencode(params_dict)
    url = "{}/{}?{}".format(config["host"], path, params)
    request = urllib.request.Request(url, file_contents, headers=config["headers"])

    try:
        response = urllib.request.urlopen(request)
    except urllib.error.HTTPError as e:
        raise ValueError(e.read().decode("utf-8"))

    job_id = response.read().decode("utf-8")
    return job_id


def cancel(config: Config, path: str, job_id):
    path = path.format(job_id)
    url = "{}/{}".format(config["host"], path)

    request = urllib.request.Request(
        url, headers=config["headers"], method="POST", data=None
    )

    try:
        response = urllib.request.urlopen(request)
        response.read()
    except urllib.error.HTTPError as e:
        raise ValueError(e.read().decode("utf-8"))


def validate(config: Config, path: str, file, organization=None):
    with open(file, "r") as ipl_file:
        content = ipl_file.read()
        file_contents = content.encode("utf-8")

    params_dict = {}
    if organization is not None:
        params_dict["organization-id"] = organization

    params = urllib.parse.urlencode(params_dict)
    url = "{}/{}?{}".format(config["host"], path, params)
    request = urllib.request.Request(url, file_contents, headers=config["headers"])

    try:
        response = urllib.request.urlopen(request)
    except urllib.error.HTTPError as e:
        raise ValueError(e.read().decode("utf-8"))

    resp = json.loads(response.read())
    return resp



# coding: utf-8

"""
    imandra-http-server

    Query Imandra via HTTP. See also https://github.com/aestheticIntegration/bs-imandra-client for a sample client implementation and OCaml API types.  # noqa: E501

    The version of the OpenAPI document: 0.0.1
    Generated by OpenAPI Generator (https://openapi-generator.tech)

    Do not edit the class manually.
"""


import unittest

import imandra_http_api_client
from imandra_http_api_client.api.default_api import DefaultApi  # noqa: E501
from imandra_http_api_client.rest import ApiException


class TestDefaultApi(unittest.TestCase):
    """DefaultApi unit test stubs"""

    def setUp(self):
        self.api = imandra_http_api_client.api.default_api.DefaultApi()  # noqa: E501

    def tearDown(self):
        pass

    def test_eval(self):
        """Test case for eval

        Evaluate a string of source code  # noqa: E501
        """
        pass

    def test_get_history(self):
        """Test case for get_history

        Get a text version of the Logic History  # noqa: E501
        """
        pass

    def test_get_status(self):
        """Test case for get_status

        Get server status  # noqa: E501
        """
        pass

    def test_instance_by_name(self):
        """Test case for instance_by_name

        Search for an instance from a property (boolean function) by name. (The name should exist from a previous call to `/eval`).  This is the more 'primitive' call - calls to /*/by-src are for convenience and use Imandra_util.gensym to assign a name to your `src` block first, and then act by_name, so give this endpoint a try if you're having issues with /*/by-src.   # noqa: E501
        """
        pass

    def test_instance_by_src(self):
        """Test case for instance_by_src

        Obtain an instance from a property (boolean function) from lambda expression source  # noqa: E501
        """
        pass

    def test_reset(self):
        """Test case for reset

        Reset Imandra internal state  # noqa: E501
        """
        pass

    def test_shutdown(self):
        """Test case for shutdown

        Shutdown the server  # noqa: E501
        """
        pass

    def test_verify_by_name(self):
        """Test case for verify_by_name

        Verify a property (boolean function) by name. (The name should exist from a previous call to `/eval`).  This is the more 'primitive' call - calls to /*/by-src are for convenience and use Imandra_util.gensym to assign a name to your `src` block first, and then act by_name, so give this endpoint a try if you're having issues with /*/by-src.   # noqa: E501
        """
        pass

    def test_verify_by_src(self):
        """Test case for verify_by_src

        Verify a property (boolean function) from lambda expression source  # noqa: E501
        """
        pass


if __name__ == '__main__':
    unittest.main()



# coding: utf-8

"""
    imandra-http-server

    Query Imandra via HTTP. See also https://github.com/aestheticIntegration/bs-imandra-client for a sample client implementation and OCaml API types.  # noqa: E501

    The version of the OpenAPI document: 0.0.1
    Generated by OpenAPI Generator (https://openapi-generator.tech)

    Do not edit the class manually.
"""


import unittest
import datetime

import imandra_http_api_client
from imandra_http_api_client.models.response_model import ResponseModel  # noqa: E501
from imandra_http_api_client.rest import ApiException

class TestResponseModel(unittest.TestCase):
    """ResponseModel unit test stubs"""

    def setUp(self):
        pass

    def tearDown(self):
        pass

    def make_instance(self, include_optional):
        """Test ResponseModel
            include_option is a boolean, when False only required
            params are included, when True both required and
            optional params are included """
        # uncomment below to create an instance of `ResponseModel`
        """
        model = imandra_http_api_client.models.response_model.ResponseModel()  # noqa: E501
        if include_optional :
            return ResponseModel(
                syntax = 'iml', 
                src = ''
            )
        else :
            return ResponseModel(
        )
        """

    def testResponseModel(self):
        """Test ResponseModel"""
        # inst_req_only = self.make_instance(include_optional=False)
        # inst_req_and_optional = self.make_instance(include_optional=True)

if __name__ == '__main__':
    unittest.main()



from . import ipl_common


def _mk_config(auth) -> ipl_common.Config:
    return {
        "host": auth.imandra_web_host,
        "headers": {"X-Auth": auth.token},
    }


def status(auth, job_id):
    config = _mk_config(auth)
    path = "api/ipl/jobs/{}/status"
    return ipl_common.status(config, path, job_id)


def wait(auth, job_id, interval=10):
    config = _mk_config(auth)
    path = "api/ipl/jobs/{}/status"
    return ipl_common.wait(config, path, job_id, interval)


def decompose(auth, file, testgen_lang, organization, callback, doc_gen, parent_job_id):
    config = _mk_config(auth)
    path = "api/ipl/decompose"
    return ipl_common.decompose(
        config, path, file, testgen_lang, organization, callback, doc_gen, parent_job_id
    )


def data(auth, job_id, file=None):
    config = _mk_config(auth)
    path = "api/ipl/jobs/{}/data"
    return ipl_common.data(config, path, job_id, file)


def simulator(auth, file):
    config = _mk_config(auth)
    path = "simulator/create"
    return ipl_common.simulator(config, auth.zone, path, file)


def list_jobs(auth, limit=10, job_type=None):
    config = _mk_config(auth)
    path = "api/ipl/jobs"
    return ipl_common.list_jobs(config, path, limit, job_type)


def unsat_analysis(auth, file, organization, callback):
    config = _mk_config(auth)
    path = "api/ipl-unsat-analysis/jobs"
    return ipl_common.unsat_analysis(config, path, file, organization, callback)


def log_analysis_builder(
    auth, file, organization=None, callback=None, decomp_job_id=None
):
    config = _mk_config(auth)
    path = "api/ipl-log-analysis-builder/jobs"
    return ipl_common.log_analysis_builder(
        config, path, file, organization, callback, decomp_job_id
    )


def cancel(auth, job_id):
    config = _mk_config(auth)
    path = "api/ipl/jobs/{}/cancel"
    return ipl_common.cancel(config, path, job_id)



import requests

from ... import auth


class ApiError(Exception):
    pass


class Client:
    _class_reasoner = None

    def __init__(
        self, reasoner=None, api_key=None, scheme=None, host=None, api_version=None
    ):
        self._reasoner = reasoner or self._class_reasoner
        if self._reasoner is None:
            raise ValueError("Please provide a reasoner")
        self._config = auth.Config(
            api_key=api_key, scheme=scheme, host=host, api_version=api_version
        )

    def eval(self, input: str, config=None):
        headers = self._config.get_headers()
        base_url = self._config.get_url()
        url = f"{base_url}/reasoners/{self._reasoner}/eval"

        json = {"input": input}
        if config:
            json["config"] = config

        response = requests.post(url, headers=headers, json=json)

        if response.status_code != 200:
            raise ApiError(response)

        return response.json()



import inspect
import sys

from . import auth

try:
    from imandrax_api import (  # pyright: ignore [reportMissingImports]
        Client as BaseClient,
    )
except ModuleNotFoundError as err:
    note = """
        Install imandra with the optional 'core' dependency to enable imandra.core:

            pip install 'imandra[core]>=2.0.0'
    """
    err.msg += "\n\n" + inspect.cleandoc(note)
    v = sys.version_info
    if (v.major, v.minor) < (3, 12):
        note = f"""
        Note that imandrax_api requires Python >= 3.12. You are using Python {v.major}.{v.minor}.{v.micro}.
        """
        err.msg += "\n\n" + inspect.cleandoc(note)
    raise


class Client(BaseClient):
    def __init__(self, api_key=None, scheme=None, host=None, **kwargs):
        config = auth.Config(
            api_key=api_key, scheme=scheme, host=host, api_version="internal"
        )
        kwargs["url"] = config.get_url() + "/imandrax"
        kwargs["auth_token"] = config.get_api_key()
        super().__init__(**kwargs)
